from __future__ import annotations

from typing import Iterator

from ...canonical_models import JiraChangelogEvent
from ...errors import SerializationError
from ..client import JiraRestClient
from ..gen import jira_api as api
from ..mappers.jira_changelog import map_changelog_event


def iter_issue_changelog_via_rest(
    client: JiraRestClient,
    *,
    issue_key: str,
    page_size: int = 100,
) -> Iterator[JiraChangelogEvent]:
    issue_key_clean = (issue_key or "").strip()
    if not issue_key_clean:
        raise ValueError("issue_key is required")
    if page_size <= 0:
        raise ValueError("page_size must be > 0")

    start_at = 0
    seen_start_at: set[int] = set()

    while True:
        if start_at in seen_start_at:
            raise SerializationError("Pagination startAt repeated; aborting to prevent infinite loop")
        seen_start_at.add(start_at)

        payload = client.get_json(
            f"/rest/api/3/issue/{issue_key_clean}/changelog",
            params={"startAt": start_at, "maxResults": page_size},
        )
        page = api.PageBeanChangelog.from_dict(payload, "data")
        values = page.values

        for item in values:
            yield map_changelog_event(issue_key=issue_key_clean, changelog=item)

        has_is_last = isinstance(page.is_last, bool)
        if has_is_last and page.is_last:
            break

        has_total = isinstance(page.total, int) and page.total >= 0
        if has_total:
            if start_at + len(values) >= page.total:
                break
        else:
            if len(values) < page_size:
                break

        if len(values) == 0:
            if has_is_last and not page.is_last:
                raise SerializationError(
                    "Received empty page with isLast=false; cannot continue pagination"
                )
            break
        start_at += len(values)



from __future__ import annotations

from typing import Optional

from ...canonical_models import JiraIssue, JiraUser
from ..gen import jira_issues_api as api


def _require_non_empty(value: Optional[str], path: str) -> str:
    if not isinstance(value, str) or not value.strip():
        raise ValueError(f"{path} is required")
    return value.strip()


def _map_user(user: Optional[api.JiraUser], path: str) -> Optional[JiraUser]:
    if user is None:
        return None
    account_id = _require_non_empty(user.account_id, f"{path}.accountId")
    display_name = _require_non_empty(user.name, f"{path}.name")
    return JiraUser(account_id=account_id, display_name=display_name, email=None)


def map_issue(*, cloud_id: str, issue: api.JiraIssueNode) -> JiraIssue:
    cloud_id_clean = (cloud_id or "").strip()
    if not cloud_id_clean:
        raise ValueError("cloud_id is required")
    if issue is None:
        raise ValueError("issue is required")

    issue_key = _require_non_empty(issue.key, "issue.key")
    issue_type = _require_non_empty(issue.issue_type.name, "issue.issueType.name")
    status = _require_non_empty(issue.status.name, "issue.status.name")

    project = issue.project_field.project
    project_key = _require_non_empty(project.key, "issue.projectField.project.key")
    project_cloud_id = _require_non_empty(project.cloud_id, "issue.projectField.project.cloudId")
    if project_cloud_id != cloud_id_clean:
        raise ValueError("issue.projectField.project.cloudId does not match cloud_id")

    created_at = _require_non_empty(issue.created_field.date_time, "issue.createdField.dateTime")
    updated_at = _require_non_empty(issue.updated_field.date_time, "issue.updatedField.dateTime")

    resolved_at = None
    if issue.resolution_date_field is not None and issue.resolution_date_field.date_time:
        resolved_at = issue.resolution_date_field.date_time.strip()

    assignee_user = (
        issue.assignee_field.user if issue.assignee_field is not None else None
    )
    reporter_user = issue.reporter

    return JiraIssue(
        cloud_id=project_cloud_id,
        key=issue_key,
        project_key=project_key,
        issue_type=issue_type,
        status=status,
        created_at=created_at,
        updated_at=updated_at,
        resolved_at=resolved_at,
        assignee=_map_user(assignee_user, "issue.assigneeField.user"),
        reporter=_map_user(reporter_user, "issue.reporter"),
        labels=[],
        components=[],
        story_points=None,
        sprint_ids=[],
    )



# Code generated by python/tools/generate_jira_worklog_models.py. DO NOT EDIT.
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from atlassian.errors import SerializationError

PAGEINFO_HAS_END_CURSOR = True
WORKLOGS_EDGE_HAS_CURSOR = True

JIRA_ISSUE_WORKLOGS_PAGE_QUERY = """query JiraIssueWorklogsPage(
  $cloudId: ID!,
  $key: String!,
  $first: Int!,
  $after: String
) {
  issue: issueByKey(key: $key, cloudId: $cloudId) {
    worklogs(first: $first, after: $after) {
      pageInfo { hasNextPage endCursor }
      edges {
        cursor
        node {
          worklogId
          author { accountId name }
          timeSpent { timeInSeconds }
          created
          updated
          startDate
        }
      }
    }
  }
}
"""


def _expect_dict(obj: Any, path: str) -> Dict[str, Any]:
    if not isinstance(obj, dict):
        raise SerializationError(f"Expected object at {path}")
    return obj


def _expect_list(obj: Any, path: str) -> List[Any]:
    if not isinstance(obj, list):
        raise SerializationError(f"Expected list at {path}")
    return obj


def _expect_str(obj: Any, path: str) -> str:
    if not isinstance(obj, str):
        raise SerializationError(f"Expected string at {path}")
    return obj


def _expect_optional_str(obj: Any, path: str) -> Optional[str]:
    if obj is None:
        return None
    if not isinstance(obj, str):
        raise SerializationError(f"Expected string at {path}")
    if not obj:
        raise SerializationError(f"Expected non-empty string at {path}")
    return obj


def _expect_int(obj: Any, path: str) -> int:
    if not isinstance(obj, int) or isinstance(obj, bool):
        raise SerializationError(f"Expected integer at {path}")
    return obj


@dataclass(frozen=True)
class PageInfo:
    has_next_page: bool
    end_cursor: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any, path: str) -> "PageInfo":
        raw = _expect_dict(obj, path)
        has_next = raw.get("hasNextPage")
        if not isinstance(has_next, bool):
            raise SerializationError(f"Expected boolean at {path}.hasNextPage")
        end_cursor: Optional[str] = None
        if PAGEINFO_HAS_END_CURSOR:
            value = raw.get("endCursor")
            if value is not None:
                end_cursor = _expect_str(value, f"{path}.endCursor")
        return PageInfo(has_next_page=has_next, end_cursor=end_cursor)


@dataclass(frozen=True)
class JiraUser:
    account_id: str
    name: str

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraUser":
        raw = _expect_dict(obj, path)
        return JiraUser(
            account_id=_expect_str(raw.get("accountId"), f"{path}.accountId"),
            name=_expect_str(raw.get("name"), f"{path}.name"),
        )


@dataclass(frozen=True)
class JiraEstimate:
    time_in_seconds: Optional[int]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraEstimate":
        raw = _expect_dict(obj, path)
        value = raw.get("timeInSeconds")
        if value is None:
            return JiraEstimate(time_in_seconds=None)
        return JiraEstimate(time_in_seconds=_expect_int(value, f"{path}.timeInSeconds"))


@dataclass(frozen=True)
class JiraWorklogNode:
    worklog_id: str
    author: Optional[JiraUser]
    time_spent: JiraEstimate
    created: str
    updated: Optional[str]
    started: Optional[str]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraWorklogNode":
        raw = _expect_dict(obj, path)
        author_raw = raw.get("author")
        return JiraWorklogNode(
            worklog_id=_expect_str(raw.get("worklogId"), f"{path}.worklogId"),
            author=JiraUser.from_dict(author_raw, f"{path}.author") if author_raw is not None else None,
            time_spent=JiraEstimate.from_dict(raw.get("timeSpent"), f"{path}.timeSpent"),
            created=_expect_str(raw.get("created"), f"{path}.created"),
            updated=_expect_optional_str(raw.get("updated"), f"{path}.updated"),
            started=_expect_optional_str(raw.get("startDate"), f"{path}.startDate"),
        )


@dataclass(frozen=True)
class JiraWorklogEdge:
    cursor: Optional[str]
    node: JiraWorklogNode

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraWorklogEdge":
        raw = _expect_dict(obj, path)
        cursor: Optional[str] = None
        if WORKLOGS_EDGE_HAS_CURSOR:
            value = raw.get("cursor")
            if value is not None:
                cursor = _expect_str(value, f"{path}.cursor")
        node = JiraWorklogNode.from_dict(raw.get("node"), f"{path}.node")
        return JiraWorklogEdge(cursor=cursor, node=node)


@dataclass(frozen=True)
class JiraWorklogConnection:
    page_info: PageInfo
    edges: List[JiraWorklogEdge]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraWorklogConnection":
        raw = _expect_dict(obj, path)
        page_info = PageInfo.from_dict(raw.get("pageInfo"), f"{path}.pageInfo")
        edges_list = _expect_list(raw.get("edges"), f"{path}.edges")
        edges = [
            JiraWorklogEdge.from_dict(item, f"{path}.edges[{idx}]")
            for idx, item in enumerate(edges_list)
        ]
        return JiraWorklogConnection(page_info=page_info, edges=edges)


def parse_issue_worklogs_page(data: Any) -> JiraWorklogConnection:
    root = _expect_dict(data, "data")
    issue = root.get("issue")
    if issue is None:
        raise SerializationError("Missing data.issue")
    issue_obj = _expect_dict(issue, "data.issue")
    worklogs = issue_obj.get("worklogs")
    if worklogs is None:
        raise SerializationError("Missing data.issue.worklogs")
    return JiraWorklogConnection.from_dict(worklogs, "data.issue.worklogs")



# Code generated by python/tools/generate_jira_project_models.py. DO NOT EDIT.
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence

from atlassian.errors import SerializationError

PAGEINFO_HAS_END_CURSOR = True
PROJECTS_EDGE_HAS_CURSOR = True
OPSGENIE_EDGE_HAS_CURSOR = True
PROJECT_HAS_ID = True
REFETCH_STRATEGY = "node"

JIRA_PROJECTS_PAGE_QUERY_TEMPLATE = """query JiraProjectsPage(
  $cloudId: ID!,
  $first: Int!,
  $after: String,
  $opsFirst: Int!
) {
  jira {
    projects: allJiraProjects(
      cloudId: $cloudId,
      filter: { types: [__PROJECT_TYPES__] },
      first: $first,
      after: $after
    ) {
      pageInfo { hasNextPage endCursor }
      edges {
        cursor
        node {
          id
          key
          name
          opsgenieTeams: opsgenieTeamsAvailableToLinkWith(first: $opsFirst) {
            pageInfo { hasNextPage endCursor }
            edges {
              cursor
              node {
                id
                name
              }
            }
          }
        }
      }
    }
  }
}
"""

JIRA_PROJECT_OPSGENIE_TEAMS_PAGE_QUERY = """query JiraProjectOpsgenieTeamsPage(
  $projectId: ID!,
  $first: Int!,
  $after: String
) {
  project: node(id: $projectId) {
    ... on JiraProject {
      opsgenieTeams: opsgenieTeamsAvailableToLinkWith(first: $first, after: $after) {
        pageInfo { hasNextPage endCursor }
        edges {
          cursor
          node {
            id
            name
          }
        }
      }
    }
  }
}
"""


def build_jira_projects_page_query(project_types: Sequence[str]) -> str:
    if not project_types:
        raise ValueError("project_types must be non-empty")
    cleaned: List[str] = []
    for raw in project_types:
        if not isinstance(raw, str):
            raise ValueError("project_types must be strings")
        value = raw.strip().upper()
        if not value or not value.replace("_", "").isalnum() or not value[0].isalpha():
            raise ValueError(f"invalid Jira project type enum token: {raw!r}")
        cleaned.append(value)
    return JIRA_PROJECTS_PAGE_QUERY_TEMPLATE.replace("__PROJECT_TYPES__", ", ".join(cleaned))


def _expect_dict(obj: Any, path: str) -> Dict[str, Any]:
    if not isinstance(obj, dict):
        raise SerializationError(f"Expected object at {path}")
    return obj


def _expect_list(obj: Any, path: str) -> List[Any]:
    if not isinstance(obj, list):
        raise SerializationError(f"Expected list at {path}")
    return obj


def _expect_str(obj: Any, path: str) -> str:
    if not isinstance(obj, str):
        raise SerializationError(f"Expected string at {path}")
    return obj


def _expect_bool(obj: Any, path: str) -> bool:
    if not isinstance(obj, bool):
        raise SerializationError(f"Expected boolean at {path}")
    return obj


@dataclass(frozen=True)
class PageInfo:
    has_next_page: bool
    end_cursor: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any, path: str) -> "PageInfo":
        raw = _expect_dict(obj, path)
        has_next = _expect_bool(raw.get("hasNextPage"), f"{path}.hasNextPage")
        end_cursor: Optional[str] = None
        if PAGEINFO_HAS_END_CURSOR:
            value = raw.get("endCursor")
            if value is not None:
                end_cursor = _expect_str(value, f"{path}.endCursor")
        return PageInfo(has_next_page=has_next, end_cursor=end_cursor)


@dataclass(frozen=True)
class OpsgenieTeamNode:
    id: str
    name: str

    @staticmethod
    def from_dict(obj: Any, path: str) -> "OpsgenieTeamNode":
        raw = _expect_dict(obj, path)
        return OpsgenieTeamNode(
            id=_expect_str(raw.get("id"), f"{path}.id"),
            name=_expect_str(raw.get("name"), f"{path}.name"),
        )


@dataclass(frozen=True)
class OpsgenieTeamEdge:
    cursor: Optional[str]
    node: OpsgenieTeamNode

    @staticmethod
    def from_dict(obj: Any, path: str) -> "OpsgenieTeamEdge":
        raw = _expect_dict(obj, path)
        cursor: Optional[str] = None
        if OPSGENIE_EDGE_HAS_CURSOR:
            value = raw.get("cursor")
            if value is not None:
                cursor = _expect_str(value, f"{path}.cursor")
        node = OpsgenieTeamNode.from_dict(raw.get("node"), f"{path}.node")
        return OpsgenieTeamEdge(cursor=cursor, node=node)


@dataclass(frozen=True)
class OpsgenieTeamsConnection:
    page_info: PageInfo
    edges: List[OpsgenieTeamEdge]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "OpsgenieTeamsConnection":
        raw = _expect_dict(obj, path)
        page_info = PageInfo.from_dict(raw.get("pageInfo"), f"{path}.pageInfo")
        edges_list = _expect_list(raw.get("edges"), f"{path}.edges")
        edges = [
            OpsgenieTeamEdge.from_dict(item, f"{path}.edges[{idx}]")
            for idx, item in enumerate(edges_list)
        ]
        return OpsgenieTeamsConnection(page_info=page_info, edges=edges)


@dataclass(frozen=True)
class JiraProjectNode:
    id: Optional[str]
    key: str
    name: str
    opsgenie_teams: OpsgenieTeamsConnection

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraProjectNode":
        raw = _expect_dict(obj, path)
        project_id: Optional[str] = None
        if PROJECT_HAS_ID:
            value = raw.get("id")
            if value is not None:
                project_id = _expect_str(value, f"{path}.id")
        return JiraProjectNode(
            id=project_id,
            key=_expect_str(raw.get("key"), f"{path}.key"),
            name=_expect_str(raw.get("name"), f"{path}.name"),
            opsgenie_teams=OpsgenieTeamsConnection.from_dict(
                raw.get("opsgenieTeams"), f"{path}.opsgenieTeams"
            ),
        )


@dataclass(frozen=True)
class JiraProjectEdge:
    cursor: Optional[str]
    node: JiraProjectNode

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraProjectEdge":
        raw = _expect_dict(obj, path)
        cursor: Optional[str] = None
        if PROJECTS_EDGE_HAS_CURSOR:
            value = raw.get("cursor")
            if value is not None:
                cursor = _expect_str(value, f"{path}.cursor")
        node = JiraProjectNode.from_dict(raw.get("node"), f"{path}.node")
        return JiraProjectEdge(cursor=cursor, node=node)


@dataclass(frozen=True)
class JiraProjectsConnection:
    page_info: PageInfo
    edges: List[JiraProjectEdge]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraProjectsConnection":
        raw = _expect_dict(obj, path)
        page_info = PageInfo.from_dict(raw.get("pageInfo"), f"{path}.pageInfo")
        edges_list = _expect_list(raw.get("edges"), f"{path}.edges")
        edges = [
            JiraProjectEdge.from_dict(item, f"{path}.edges[{idx}]")
            for idx, item in enumerate(edges_list)
        ]
        return JiraProjectsConnection(page_info=page_info, edges=edges)


@dataclass(frozen=True)
class JiraProjectsPageData:
    projects: JiraProjectsConnection

    @staticmethod
    def from_dict(obj: Any, path: str = "data.jira") -> "JiraProjectsPageData":
        raw = _expect_dict(obj, path)
        return JiraProjectsPageData(
            projects=JiraProjectsConnection.from_dict(
                raw.get("projects"), f"{path}.projects"
            ),
        )


def parse_jira_projects_page(data: Any) -> JiraProjectsPageData:
    root = _expect_dict(data, "data")
    jira = root.get("jira")
    if jira is None:
        raise SerializationError("Missing data.jira")
    return JiraProjectsPageData.from_dict(jira, "data.jira")


def parse_project_opsgenie_teams(data: Any) -> OpsgenieTeamsConnection:
    root = _expect_dict(data, "data")
    if REFETCH_STRATEGY == "node":
        project = root.get("project")
        if project is None:
            raise SerializationError("Missing data.project")
        project_obj = _expect_dict(project, "data.project")
        inner = project_obj.get("opsgenieTeams")
        if inner is None:
            raise SerializationError("Missing data.project.opsgenieTeams")
        return OpsgenieTeamsConnection.from_dict(inner, "data.project.opsgenieTeams")

    jira = root.get("jira")
    if jira is None:
        raise SerializationError("Missing data.jira")
    jira_obj = _expect_dict(jira, "data.jira")
    project = jira_obj.get("project")
    if project is None:
        raise SerializationError("Missing data.jira.project")
    project_obj = _expect_dict(project, "data.jira.project")
    inner = project_obj.get("opsgenieTeams")
    if inner is None:
        raise SerializationError("Missing data.jira.project.opsgenieTeams")
    return OpsgenieTeamsConnection.from_dict(inner, "data.jira.project.opsgenieTeams")



import httpx

from atlassian.auth import OAuthBearerAuth
from atlassian.rest.api.jira_changelog import iter_issue_changelog_via_rest
from atlassian.rest.client import JiraRestClient


def test_jira_rest_changelog_pagination_and_mapping():
    def handler(request: httpx.Request) -> httpx.Response:
        assert request.url.path.endswith("/rest/api/3/issue/A-1/changelog")
        start_at = int(request.url.params.get("startAt", "0"))
        if start_at == 0:
            return httpx.Response(
                200,
                json={
                    "startAt": 0,
                    "maxResults": 1,
                    "total": 2,
                    "isLast": False,
                    "values": [
                        {
                            "id": "100",
                            "created": "2021-01-02T00:00:00.000+0000",
                            "author": {"accountId": "u1", "displayName": "User 1"},
                            "items": [
                                {"field": "status", "fromString": "To Do", "toString": "In Progress"}
                            ],
                        }
                    ],
                },
                request=request,
            )
        if start_at == 1:
            return httpx.Response(
                200,
                json={
                    "startAt": 1,
                    "maxResults": 1,
                    "total": 2,
                    "isLast": True,
                    "values": [
                        {
                            "id": "101",
                            "created": "2021-01-03T00:00:00.000+0000",
                            "items": [{"field": "assignee", "from": "u1", "to": "u2"}],
                        }
                    ],
                },
                request=request,
            )
        raise AssertionError(f"unexpected startAt={start_at}")

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = JiraRestClient(
            "https://api.atlassian.com/ex/jira/cloud-123",
            auth=OAuthBearerAuth(lambda: "token"),
            http_client=http_client,
        )
        events = list(iter_issue_changelog_via_rest(client, issue_key="A-1", page_size=1))

    assert len(events) == 2
    assert events[0].issue_key == "A-1"
    assert events[0].event_id == "100"
    assert events[0].author and events[0].author.account_id == "u1"
    assert events[0].items[0].field == "status"
    assert events[1].author is None



# Code generated by python/tools/generate_jira_sprint_models.py. DO NOT EDIT.
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional

from atlassian.errors import SerializationError

JIRA_SPRINT_BY_ID_QUERY = """query JiraSprintById(
  $id: ID!
) {
  sprintById(id: $id) {
    sprintId
    name
    state
    startDate
    endDate
    completionDate
  }
}
"""


def _expect_dict(obj: Any, path: str) -> Dict[str, Any]:
    if not isinstance(obj, dict):
        raise SerializationError(f"Expected object at {path}")
    return obj


def _expect_str(obj: Any, path: str) -> str:
    if not isinstance(obj, str):
        raise SerializationError(f"Expected string at {path}")
    return obj


def _expect_optional_str(obj: Any, path: str) -> Optional[str]:
    if obj is None:
        return None
    if not isinstance(obj, str):
        raise SerializationError(f"Expected string at {path}")
    if not obj:
        raise SerializationError(f"Expected non-empty string at {path}")
    return obj


@dataclass(frozen=True)
class JiraSprintNode:
    sprint_id: str
    name: Optional[str]
    state: Optional[str]
    start_date: Optional[str]
    end_date: Optional[str]
    completion_date: Optional[str]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraSprintNode":
        raw = _expect_dict(obj, path)
        return JiraSprintNode(
            sprint_id=_expect_str(raw.get("sprintId"), f"{path}.sprintId"),
            name=_expect_optional_str(raw.get("name"), f"{path}.name"),
            state=_expect_optional_str(raw.get("state"), f"{path}.state"),
            start_date=_expect_optional_str(raw.get("startDate"), f"{path}.startDate"),
            end_date=_expect_optional_str(raw.get("endDate"), f"{path}.endDate"),
            completion_date=_expect_optional_str(raw.get("completionDate"), f"{path}.completionDate"),
        )


def parse_jira_sprint_by_id(data: Any) -> JiraSprintNode:
    root = _expect_dict(data, "data")
    sprint = root.get("sprintById")
    if sprint is None:
        raise SerializationError("Missing data.sprintById")
    return JiraSprintNode.from_dict(sprint, "data.sprintById")



from __future__ import annotations

from typing import List, Optional

from ...canonical_models import JiraChangelogEvent, JiraChangelogItem, JiraUser
from ..gen.jira_api import Changelog, ChangeDetails, UserDetails


def _require_non_empty(value: Optional[str], path: str) -> str:
    if not isinstance(value, str) or not value.strip():
        raise ValueError(f"{path} is required")
    return value.strip()


def _map_user(user: Optional[UserDetails], path: str) -> Optional[JiraUser]:
    if user is None:
        return None
    account_id = user.account_id
    display_name = user.display_name
    email = user.email_address
    return JiraUser(
        account_id=_require_non_empty(account_id, f"{path}.accountId"),
        display_name=_require_non_empty(display_name, f"{path}.displayName"),
        email=email.strip() if isinstance(email, str) and email.strip() else None,
    )


def _map_item(item: ChangeDetails, path: str) -> JiraChangelogItem:
    field = _require_non_empty(item.field, f"{path}.field")
    from_value = item.from_value.strip() if isinstance(item.from_value, str) and item.from_value.strip() else None
    to_value = item.to_value.strip() if isinstance(item.to_value, str) and item.to_value.strip() else None
    from_string = item.from_string.strip() if isinstance(item.from_string, str) and item.from_string.strip() else None
    to_string = item.to_string.strip() if isinstance(item.to_string, str) and item.to_string.strip() else None
    return JiraChangelogItem(
        field=field,
        from_value=from_value,
        to_value=to_value,
        from_string=from_string,
        to_string=to_string,
    )


def map_changelog_event(*, issue_key: str, changelog: Changelog) -> JiraChangelogEvent:
    issue_key_clean = (issue_key or "").strip()
    if not issue_key_clean:
        raise ValueError("issue_key is required")
    if changelog is None:
        raise ValueError("changelog is required")

    event_id = _require_non_empty(changelog.id, "changelog.id")
    created_at = _require_non_empty(changelog.created, "changelog.created")

    items: List[JiraChangelogItem] = [
        _map_item(item, f"changelog.items[{idx}]")
        for idx, item in enumerate(changelog.items)
    ]
    if not items:
        raise ValueError("changelog.items is required")

    return JiraChangelogEvent(
        issue_key=issue_key_clean,
        event_id=event_id,
        created_at=created_at,
        items=items,
        author=_map_user(changelog.author, "changelog.author"),
    )



import json

import httpx

from atlassian.auth import OAuthBearerAuth
from atlassian.graph.api.jira_worklogs import iter_issue_worklogs_via_graphql
from atlassian.graph.client import GraphQLClient


def test_jira_worklogs_graphql_pagination():
    calls = []

    def handler(request: httpx.Request) -> httpx.Response:
        payload = json.loads(request.content.decode("utf-8"))
        calls.append(payload)
        after = (payload.get("variables") or {}).get("after")
        if after in (None, ""):
            data = {
                "data": {
                    "issue": {
                        "worklogs": {
                            "pageInfo": {"hasNextPage": True, "endCursor": "c1"},
                            "edges": [
                                {
                                    "cursor": "e1",
                                    "node": {
                                        "worklogId": "w1",
                                        "author": {"accountId": "u1", "name": "User One"},
                                        "timeSpent": {"timeInSeconds": 60},
                                        "created": "2021-01-01T00:00:00Z",
                                        "updated": "2021-01-01T01:00:00Z",
                                        "startDate": "2021-01-01T00:00:00Z",
                                    },
                                }
                            ],
                        }
                    }
                }
            }
        else:
            data = {
                "data": {
                    "issue": {
                        "worklogs": {
                            "pageInfo": {"hasNextPage": False, "endCursor": None},
                            "edges": [
                                {
                                    "cursor": "e2",
                                    "node": {
                                        "worklogId": "w2",
                                        "author": {"accountId": "u2", "name": "User Two"},
                                        "timeSpent": {"timeInSeconds": 120},
                                        "created": "2021-01-02T00:00:00Z",
                                        "updated": "2021-01-02T01:00:00Z",
                                        "startDate": "2021-01-02T00:00:00Z",
                                    },
                                }
                            ],
                        }
                    }
                }
            }
        return httpx.Response(200, json=data, request=request)

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            http_client=http_client,
        )
        worklogs = list(
            iter_issue_worklogs_via_graphql(
                client,
                cloud_id="cloud-123",
                issue_key="A-1",
                page_size=1,
            )
        )

    assert [wl.worklog_id for wl in worklogs] == ["w1", "w2"]
    assert worklogs[0].author and worklogs[0].author.account_id == "u1"
    assert worklogs[1].time_spent_seconds == 120
    assert len(calls) == 2



from __future__ import annotations

from typing import Dict, Iterator, Optional, Union

from ...canonical_models import JiraSprint
from ...errors import SerializationError
from ..client import JiraRestClient
from ..gen.jira_agile_api import SprintPage
from ..mappers.jira_sprints import map_sprint


def iter_board_sprints_via_rest(
    client: JiraRestClient,
    *,
    board_id: int,
    state: Optional[str] = None,
    page_size: int = 50,
) -> Iterator[JiraSprint]:
    """Iterate over sprints for a Jira Agile board.

    Args:
        client: JiraRestClient instance
        board_id: The ID of the Jira Agile board
        state: Optional filter by sprint state (future, active, closed)
        page_size: Number of sprints per page (default: 50)

    Yields:
        JiraSprint: Canonical sprint objects

    Raises:
        ValueError: If board_id is invalid or page_size is <= 0
        SerializationError: If pagination loops
    """
    if board_id <= 0:
        raise ValueError("board_id must be a positive integer")
    if page_size <= 0:
        raise ValueError("page_size must be > 0")

    state_clean: Optional[str] = None
    if state is not None:
        state_clean = state.strip().lower()
        if not state_clean:
            # Treat empty/whitespace-only state as no filter for consistency with Go implementation
            state_clean = None
        elif state_clean not in ("future", "active", "closed"):
            raise ValueError("state must be one of: future, active, closed")

    start_at = 0
    seen_start_at: set[int] = set()

    while True:
        if start_at in seen_start_at:
            raise SerializationError("Pagination startAt repeated; aborting to prevent infinite loop")
        seen_start_at.add(start_at)

        params: Dict[str, Union[int, str]] = {"startAt": start_at, "maxResults": page_size}
        if state_clean is not None:
            params["state"] = state_clean

        payload = client.get_json(
            f"/rest/agile/1.0/board/{board_id}/sprint",
            params=params,
        )
        page = SprintPage.from_dict(payload, "data")
        values = page.values

        for item in values:
            yield map_sprint(sprint=item)

        has_is_last = isinstance(page.is_last, bool)
        if has_is_last and page.is_last:
            break

        if len(values) < page_size:
            break

        if len(values) == 0:
            if has_is_last and not page.is_last:
                raise SerializationError(
                    "Received empty page with isLast=false; cannot continue pagination"
                )
            break
        start_at += len(values)



from __future__ import annotations

import json
import os
from typing import Optional

from ..auth import BasicApiTokenAuth, CookieAuth, OAuthBearerAuth
from ..oauth_3lo import OAuthRefreshTokenAuth


def auth_from_env():
    token = os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN")
    refresh_token = os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
    client_id = os.getenv("ATLASSIAN_CLIENT_ID")
    client_secret = os.getenv("ATLASSIAN_CLIENT_SECRET")
    email = os.getenv("ATLASSIAN_EMAIL")
    api_token = os.getenv("ATLASSIAN_API_TOKEN")
    cookies_json = os.getenv("ATLASSIAN_COOKIES_JSON")

    if refresh_token and client_id and client_secret:
        return OAuthRefreshTokenAuth(
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=refresh_token,
        )
    if token:
        if client_secret and token.strip() == client_secret.strip():
            raise ValueError(
                "ATLASSIAN_OAUTH_ACCESS_TOKEN appears to be set to ATLASSIAN_CLIENT_SECRET; "
                "set an OAuth access token (not the client secret)."
            )
        return OAuthBearerAuth(lambda: token)
    if email and api_token:
        return BasicApiTokenAuth(email, api_token)
    if cookies_json:
        try:
            cookies = json.loads(cookies_json)
        except json.JSONDecodeError:
            return None
        if isinstance(cookies, dict) and all(
            isinstance(k, str) and isinstance(v, str) for k, v in cookies.items()
        ):
            return CookieAuth(cookies)
    return None


def _derive_site_base_url_from_gql_base(gql_base_url: str) -> Optional[str]:
    candidate = (gql_base_url or "").strip().rstrip("/")
    if not candidate:
        return None
    for suffix in ("/gateway/api/graphql", "/gateway/api", "/graphql"):
        if candidate.endswith(suffix):
            stripped = candidate[: -len(suffix)]
            return stripped.rstrip("/")
    return None


def jira_rest_base_url_from_env(cloud_id: str) -> Optional[str]:
    explicit = os.getenv("ATLASSIAN_JIRA_BASE_URL")
    if explicit and explicit.strip():
        return explicit.strip().rstrip("/")

    if os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN") or os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN"):
        cloud_id_clean = (cloud_id or "").strip()
        if not cloud_id_clean:
            return None
        return f"https://api.atlassian.com/ex/jira/{cloud_id_clean}"

    gql_base = os.getenv("ATLASSIAN_GQL_BASE_URL")
    if gql_base:
        return _derive_site_base_url_from_gql_base(gql_base)
    return None



from __future__ import annotations

from typing import Any, Dict, List, Optional

from ...canonical_models import JiraIssue, JiraUser
from ..gen.jira_api import IssueBean


def _expect_dict(obj: Any, path: str) -> Dict[str, Any]:
    if not isinstance(obj, dict):
        raise ValueError(f"Expected object at {path}")
    return obj


def _expect_str(obj: Any, path: str) -> str:
    if not isinstance(obj, str):
        raise ValueError(f"Expected string at {path}")
    value = obj.strip()
    if not value:
        raise ValueError(f"Expected non-empty string at {path}")
    return value


def _expect_list(obj: Any, path: str) -> List[Any]:
    if not isinstance(obj, list):
        raise ValueError(f"Expected list at {path}")
    return obj


def _maybe_user(obj: Any, path: str) -> Optional[JiraUser]:
    if obj is None:
        return None
    raw = _expect_dict(obj, path)
    account_id = raw.get("accountId")
    display_name = raw.get("displayName")
    email = raw.get("emailAddress")
    email_value: Optional[str] = None
    if email is not None:
        if not isinstance(email, str):
            raise ValueError(f"Expected string at {path}.emailAddress")
        if email.strip():
            email_value = email.strip()
    return JiraUser(
        account_id=_expect_str(account_id, f"{path}.accountId"),
        display_name=_expect_str(display_name, f"{path}.displayName"),
        email=email_value,
    )


def _parse_story_points(fields: Dict[str, Any], field_name: Optional[str]) -> Optional[float]:
    if not field_name:
        return None
    field_key = field_name.strip()
    if not field_key:
        raise ValueError("story_points_field must be non-empty when provided")
    raw = fields.get(field_key)
    if raw is None:
        return None
    if isinstance(raw, bool):
        raise ValueError(f"issue.fields.{field_key} must be a number when present")
    if isinstance(raw, (int, float)):
        return float(raw)
    if isinstance(raw, str):
        if not raw.strip():
            return None
        try:
            return float(raw)
        except ValueError as exc:
            raise ValueError(f"issue.fields.{field_key} must be a number when present") from exc
    raise ValueError(f"issue.fields.{field_key} must be a number when present")


def _coerce_sprint_id(value: Any, path: str) -> str:
    if isinstance(value, bool):
        raise ValueError(f"{path} must be a string or integer")
    if isinstance(value, int):
        return str(value)
    if isinstance(value, float):
        if value.is_integer():
            return str(int(value))
        raise ValueError(f"{path} must be an integer")
    if isinstance(value, str):
        cleaned = value.strip()
        if not cleaned:
            raise ValueError(f"{path} must be non-empty")
        return cleaned
    raise ValueError(f"{path} must be a string or integer")


def _parse_sprint_ids(fields: Dict[str, Any], field_name: Optional[str]) -> List[str]:
    if not field_name:
        return []
    field_key = field_name.strip()
    if not field_key:
        raise ValueError("sprint_ids_field must be non-empty when provided")
    raw = fields.get(field_key)
    if raw is None:
        return []
    items = _expect_list(raw, f"issue.fields.{field_key}")
    out: List[str] = []
    for idx, item in enumerate(items):
        path = f"issue.fields.{field_key}[{idx}]"
        if isinstance(item, dict):
            sprint_id = item.get("id")
            if sprint_id is None:
                raise ValueError(f"{path}.id is required")
            out.append(_coerce_sprint_id(sprint_id, f"{path}.id"))
        else:
            out.append(_coerce_sprint_id(item, path))
    return out


def map_issue(
    *,
    cloud_id: str,
    issue: IssueBean,
    story_points_field: Optional[str] = None,
    sprint_ids_field: Optional[str] = None,
) -> JiraIssue:
    cloud_id_clean = (cloud_id or "").strip()
    if not cloud_id_clean:
        raise ValueError("cloud_id is required")
    if issue is None:
        raise ValueError("issue is required")

    issue_key = _expect_str(issue.key, "issue.key")
    fields = _expect_dict(issue.fields, "issue.fields")

    project = _expect_dict(fields.get("project"), "issue.fields.project")
    project_key = _expect_str(project.get("key"), "issue.fields.project.key")

    issuetype = _expect_dict(fields.get("issuetype"), "issue.fields.issuetype")
    issue_type = _expect_str(issuetype.get("name"), "issue.fields.issuetype.name")

    status_obj = _expect_dict(fields.get("status"), "issue.fields.status")
    status = _expect_str(status_obj.get("name"), "issue.fields.status.name")

    created_at = _expect_str(fields.get("created"), "issue.fields.created")
    updated_at = _expect_str(fields.get("updated"), "issue.fields.updated")

    resolved_at: Optional[str] = None
    resolutiondate = fields.get("resolutiondate")
    if resolutiondate is not None:
        if not isinstance(resolutiondate, str):
            raise ValueError("issue.fields.resolutiondate must be a string when present")
        if resolutiondate.strip():
            resolved_at = resolutiondate.strip()

    labels: List[str] = []
    raw_labels = fields.get("labels")
    if raw_labels is not None:
        for idx, item in enumerate(_expect_list(raw_labels, "issue.fields.labels")):
            labels.append(_expect_str(item, f"issue.fields.labels[{idx}]"))

    components: List[str] = []
    raw_components = fields.get("components")
    if raw_components is not None:
        for idx, comp in enumerate(_expect_list(raw_components, "issue.fields.components")):
            comp_obj = _expect_dict(comp, f"issue.fields.components[{idx}]")
            components.append(_expect_str(comp_obj.get("name"), f"issue.fields.components[{idx}].name"))

    assignee = _maybe_user(fields.get("assignee"), "issue.fields.assignee")
    reporter = _maybe_user(fields.get("reporter"), "issue.fields.reporter")

    story_points = _parse_story_points(fields, story_points_field)
    sprint_ids = _parse_sprint_ids(fields, sprint_ids_field)

    return JiraIssue(
        cloud_id=cloud_id_clean,
        key=issue_key,
        project_key=project_key,
        issue_type=issue_type,
        status=status,
        created_at=created_at,
        updated_at=updated_at,
        resolved_at=resolved_at,
        assignee=assignee,
        reporter=reporter,
        labels=labels,
        components=components,
        story_points=story_points,
        sprint_ids=sprint_ids,
    )



from __future__ import annotations

import json
import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple


def _add_project_to_syspath() -> None:
    project_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(project_root))


_add_project_to_syspath()

from atlassian.auth import (  # noqa: E402
    BasicApiTokenAuth,
    CookieAuth,
    OAuthBearerAuth,
)
from atlassian.oauth_3lo import OAuthRefreshTokenAuth  # noqa: E402
from atlassian.graph.schema_fetcher import fetch_schema_introspection  # noqa: E402


@dataclass(frozen=True)
class _Config:
    pageinfo_has_end_cursor: bool
    edge_has_cursor: bool


def _env_experimental_apis() -> List[str]:
    raw = os.getenv("ATLASSIAN_GQL_EXPERIMENTAL_APIS", "")
    return [part.strip() for part in raw.split(",") if part.strip()]


def _maybe_strip_quotes(raw: str) -> str:
    if len(raw) >= 2 and raw[0] == raw[-1] and raw[0] in {"'", '"'}:
        return raw[1:-1]
    return raw


def _load_env_file(path: Path) -> None:
    if not path.exists():
        return
    for line in path.read_text(encoding="utf-8").splitlines():
        stripped = line.strip()
        if not stripped or stripped.startswith("#"):
            continue
        if stripped.startswith("export "):
            stripped = stripped[len("export ") :].strip()
        if "=" not in stripped:
            continue
        key, _, value = stripped.partition("=")
        key = key.strip()
        if not key or key in os.environ:
            continue
        os.environ[key] = _maybe_strip_quotes(value.strip())


def _build_auth_from_env():
    token = os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN")
    refresh_token = os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
    client_id = os.getenv("ATLASSIAN_CLIENT_ID")
    client_secret = os.getenv("ATLASSIAN_CLIENT_SECRET")
    email = os.getenv("ATLASSIAN_EMAIL")
    api_token = os.getenv("ATLASSIAN_API_TOKEN")
    cookies_json = os.getenv("ATLASSIAN_COOKIES_JSON")

    if refresh_token and client_id and client_secret:
        return OAuthRefreshTokenAuth(
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=refresh_token,
        )
    if token:
        if client_secret and token.strip() == client_secret.strip():
            raise ValueError(
                "ATLASSIAN_OAUTH_ACCESS_TOKEN appears to be set to ATLASSIAN_CLIENT_SECRET; "
                "set an OAuth access token (not the client secret)."
            )
        return OAuthBearerAuth(lambda: token)
    if email and api_token:
        return BasicApiTokenAuth(email, api_token)
    if cookies_json:
        try:
            cookies = json.loads(cookies_json)
        except json.JSONDecodeError:
            return None
        if isinstance(cookies, dict) and all(
            isinstance(k, str) and isinstance(v, str) for k, v in cookies.items()
        ):
            return CookieAuth(cookies)
    return None


def _load_introspection(path: Path) -> Dict[str, Any]:
    raw = json.loads(path.read_text(encoding="utf-8"))
    if isinstance(raw, dict) and "data" in raw and isinstance(raw["data"], dict):
        data = raw["data"]
    else:
        data = raw
    if not isinstance(data, dict) or "__schema" not in data:
        raise RuntimeError("Introspection JSON missing data.__schema")
    schema = data["__schema"]
    if not isinstance(schema, dict):
        raise RuntimeError("Introspection JSON data.__schema is not an object")
    return schema


def _types_map(schema: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
    types = schema.get("types")
    if not isinstance(types, list):
        raise RuntimeError("Introspection JSON missing __schema.types[]")
    out: Dict[str, Dict[str, Any]] = {}
    for t in types:
        if not isinstance(t, dict):
            continue
        name = t.get("name")
        if isinstance(name, str) and name:
            out[name] = t
    return out


def _unwrap_named_type(type_ref: Dict[str, Any]) -> Tuple[Optional[str], Optional[str], Dict[str, Any]]:
    cur = type_ref
    for _ in range(16):
        if not isinstance(cur, dict):
            break
        kind = cur.get("kind")
        name = cur.get("name")
        if isinstance(name, str) and name:
            return name, kind if isinstance(kind, str) else None, cur
        nxt = cur.get("ofType")
        if nxt is None:
            break
        cur = nxt
    return None, None, {}


def _field(type_def: Dict[str, Any], name: str) -> Optional[Dict[str, Any]]:
    fields = type_def.get("fields")
    if not isinstance(fields, list):
        return None
    for f in fields:
        if isinstance(f, dict) and f.get("name") == name:
            return f
    return None


def _arg(field_def: Dict[str, Any], name: str) -> Optional[Dict[str, Any]]:
    args = field_def.get("args")
    if not isinstance(args, list):
        return None
    for a in args:
        if isinstance(a, dict) and a.get("name") == name:
            return a
    return None


def _require_field(type_def: Dict[str, Any], name: str, path: str) -> Dict[str, Any]:
    found = _field(type_def, name)
    if not found:
        raise RuntimeError(f"Missing required field {path}.{name}")
    return found


def _require_type(types: Dict[str, Dict[str, Any]], name: str) -> Dict[str, Any]:
    type_def = types.get(name)
    if not type_def:
        raise RuntimeError(f"Missing type definition: {name}")
    return type_def


def _discover_config(schema: Dict[str, Any]) -> _Config:
    types = _types_map(schema)

    query_type = schema.get("queryType")
    query_name = query_type.get("name") if isinstance(query_type, dict) else None
    if not isinstance(query_name, str) or not query_name:
        raise RuntimeError("Introspection JSON missing __schema.queryType.name")
    query_def = _require_type(types, query_name)

    issue_by_key_field = _field(query_def, "issueByKey")
    if not issue_by_key_field:
        raise RuntimeError(f"Missing required field {query_name}.issueByKey")
    for arg_name in ("key", "cloudId"):
        if not _arg(issue_by_key_field, arg_name):
            raise RuntimeError(f"Missing required arg issueByKey.{arg_name}")

    issue_type_name, _, _ = _unwrap_named_type(issue_by_key_field.get("type") or {})
    if not issue_type_name:
        raise RuntimeError("Unable to resolve type of issueByKey")
    issue_def = _require_type(types, issue_type_name)

    worklogs_field = _require_field(issue_def, "worklogs", f"type {issue_type_name}.fields")
    for arg_name in ("first", "after"):
        if not _arg(worklogs_field, arg_name):
            raise RuntimeError(f"Missing required arg issue.worklogs.{arg_name}")

    worklogs_conn_name, _, _ = _unwrap_named_type(worklogs_field.get("type") or {})
    if not worklogs_conn_name:
        raise RuntimeError("Unable to resolve worklogs connection type")
    worklogs_conn_def = _require_type(types, worklogs_conn_name)

    pageinfo_field = _require_field(worklogs_conn_def, "pageInfo", f"type {worklogs_conn_name}.fields")
    edges_field = _require_field(worklogs_conn_def, "edges", f"type {worklogs_conn_name}.fields")

    pageinfo_name, _, _ = _unwrap_named_type(pageinfo_field.get("type") or {})
    pageinfo_def = _require_type(types, pageinfo_name or "")
    _require_field(pageinfo_def, "hasNextPage", f"type {pageinfo_name}.fields")
    pageinfo_has_end_cursor = _field(pageinfo_def, "endCursor") is not None

    edges_name, _, _ = _unwrap_named_type(edges_field.get("type") or {})
    edges_def = _require_type(types, edges_name or "")
    edge_has_cursor = _field(edges_def, "cursor") is not None
    node_field = _require_field(edges_def, "node", f"type {edges_name}.fields")
    node_name, _, _ = _unwrap_named_type(node_field.get("type") or {})
    worklog_def = _require_type(types, node_name or "")

    _require_field(worklog_def, "worklogId", f"type {node_name}.fields")
    _require_field(worklog_def, "author", f"type {node_name}.fields")
    time_spent_field = _require_field(worklog_def, "timeSpent", f"type {node_name}.fields")
    _require_field(worklog_def, "created", f"type {node_name}.fields")
    _require_field(worklog_def, "updated", f"type {node_name}.fields")
    _require_field(worklog_def, "startDate", f"type {node_name}.fields")

    estimate_name, _, _ = _unwrap_named_type(time_spent_field.get("type") or {})
    estimate_def = _require_type(types, estimate_name or "")
    _require_field(estimate_def, "timeInSeconds", f"type {estimate_name}.fields")

    user_def = _require_type(types, "User")
    _require_field(user_def, "accountId", "type User.fields")
    _require_field(user_def, "name", "type User.fields")

    return _Config(
        pageinfo_has_end_cursor=pageinfo_has_end_cursor,
        edge_has_cursor=edge_has_cursor,
    )


def _render_python(cfg: _Config) -> str:
    return f"""# Code generated by python/tools/generate_jira_worklog_models.py. DO NOT EDIT.
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from atlassian.errors import SerializationError

PAGEINFO_HAS_END_CURSOR = {str(cfg.pageinfo_has_end_cursor)}
WORKLOGS_EDGE_HAS_CURSOR = {str(cfg.edge_has_cursor)}

JIRA_ISSUE_WORKLOGS_PAGE_QUERY = \"\"\"query JiraIssueWorklogsPage(
  $cloudId: ID!,
  $key: String!,
  $first: Int!,
  $after: String
) {{
  issue: issueByKey(key: $key, cloudId: $cloudId) {{
    worklogs(first: $first, after: $after) {{
      pageInfo {{ hasNextPage endCursor }}
      edges {{
        cursor
        node {{
          worklogId
          author {{ accountId name }}
          timeSpent {{ timeInSeconds }}
          created
          updated
          startDate
        }}
      }}
    }}
  }}
}}
\"\"\"


def _expect_dict(obj: Any, path: str) -> Dict[str, Any]:
    if not isinstance(obj, dict):
        raise SerializationError(f\"Expected object at {{path}}\")
    return obj


def _expect_list(obj: Any, path: str) -> List[Any]:
    if not isinstance(obj, list):
        raise SerializationError(f\"Expected list at {{path}}\")
    return obj


def _expect_str(obj: Any, path: str) -> str:
    if not isinstance(obj, str):
        raise SerializationError(f\"Expected string at {{path}}\")
    return obj


def _expect_optional_str(obj: Any, path: str) -> Optional[str]:
    if obj is None:
        return None
    if not isinstance(obj, str):
        raise SerializationError(f\"Expected string at {{path}}\")
    if not obj:
        raise SerializationError(f\"Expected non-empty string at {{path}}\")
    return obj


def _expect_int(obj: Any, path: str) -> int:
    if not isinstance(obj, int) or isinstance(obj, bool):
        raise SerializationError(f\"Expected integer at {{path}}\")
    return obj


@dataclass(frozen=True)
class PageInfo:
    has_next_page: bool
    end_cursor: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"PageInfo\":
        raw = _expect_dict(obj, path)
        has_next = raw.get(\"hasNextPage\")
        if not isinstance(has_next, bool):
            raise SerializationError(f\"Expected boolean at {{path}}.hasNextPage\")
        end_cursor: Optional[str] = None
        if PAGEINFO_HAS_END_CURSOR:
            value = raw.get(\"endCursor\")
            if value is not None:
                end_cursor = _expect_str(value, f\"{{path}}.endCursor\")
        return PageInfo(has_next_page=has_next, end_cursor=end_cursor)


@dataclass(frozen=True)
class JiraUser:
    account_id: str
    name: str

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraUser\":
        raw = _expect_dict(obj, path)
        return JiraUser(
            account_id=_expect_str(raw.get(\"accountId\"), f\"{{path}}.accountId\"),
            name=_expect_str(raw.get(\"name\"), f\"{{path}}.name\"),
        )


@dataclass(frozen=True)
class JiraEstimate:
    time_in_seconds: Optional[int]

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraEstimate\":
        raw = _expect_dict(obj, path)
        value = raw.get(\"timeInSeconds\")
        if value is None:
            return JiraEstimate(time_in_seconds=None)
        return JiraEstimate(time_in_seconds=_expect_int(value, f\"{{path}}.timeInSeconds\"))


@dataclass(frozen=True)
class JiraWorklogNode:
    worklog_id: str
    author: Optional[JiraUser]
    time_spent: JiraEstimate
    created: str
    updated: Optional[str]
    started: Optional[str]

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraWorklogNode\":
        raw = _expect_dict(obj, path)
        author_raw = raw.get(\"author\")
        return JiraWorklogNode(
            worklog_id=_expect_str(raw.get(\"worklogId\"), f\"{{path}}.worklogId\"),
            author=JiraUser.from_dict(author_raw, f\"{{path}}.author\") if author_raw is not None else None,
            time_spent=JiraEstimate.from_dict(raw.get(\"timeSpent\"), f\"{{path}}.timeSpent\"),
            created=_expect_str(raw.get(\"created\"), f\"{{path}}.created\"),
            updated=_expect_optional_str(raw.get(\"updated\"), f\"{{path}}.updated\"),
            started=_expect_optional_str(raw.get(\"startDate\"), f\"{{path}}.startDate\"),
        )


@dataclass(frozen=True)
class JiraWorklogEdge:
    cursor: Optional[str]
    node: JiraWorklogNode

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraWorklogEdge\":
        raw = _expect_dict(obj, path)
        cursor: Optional[str] = None
        if WORKLOGS_EDGE_HAS_CURSOR:
            value = raw.get(\"cursor\")
            if value is not None:
                cursor = _expect_str(value, f\"{{path}}.cursor\")
        node = JiraWorklogNode.from_dict(raw.get(\"node\"), f\"{{path}}.node\")
        return JiraWorklogEdge(cursor=cursor, node=node)


@dataclass(frozen=True)
class JiraWorklogConnection:
    page_info: PageInfo
    edges: List[JiraWorklogEdge]

    @staticmethod
    def from_dict(obj: Any, path: str) -> \"JiraWorklogConnection\":
        raw = _expect_dict(obj, path)
        page_info = PageInfo.from_dict(raw.get(\"pageInfo\"), f\"{{path}}.pageInfo\")
        edges_list = _expect_list(raw.get(\"edges\"), f\"{{path}}.edges\")
        edges = [
            JiraWorklogEdge.from_dict(item, f\"{{path}}.edges[{{idx}}]\")
            for idx, item in enumerate(edges_list)
        ]
        return JiraWorklogConnection(page_info=page_info, edges=edges)


def parse_issue_worklogs_page(data: Any) -> JiraWorklogConnection:
    root = _expect_dict(data, \"data\")
    issue = root.get(\"issue\")
    if issue is None:
        raise SerializationError(\"Missing data.issue\")
    issue_obj = _expect_dict(issue, \"data.issue\")
    worklogs = issue_obj.get(\"worklogs\")
    if worklogs is None:
        raise SerializationError(\"Missing data.issue.worklogs\")
    return JiraWorklogConnection.from_dict(worklogs, \"data.issue.worklogs\")
"""


def main() -> None:
    repo_root = Path(__file__).resolve().parents[1]
    token_file = Path(os.getenv("ATLASSIAN_OAUTH_TOKEN_FILE", repo_root / "oauth_tokens.txt"))
    _load_env_file(token_file)

    schema_path = repo_root / "graphql" / "schema.introspection.json"
    if not schema_path.exists():
        base_url = os.getenv("ATLASSIAN_GQL_BASE_URL")
        if not base_url and (
            os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN") or os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
        ):
            base_url = "https://api.atlassian.com"
        if not base_url:
            raise RuntimeError(f"Missing {schema_path} and ATLASSIAN_GQL_BASE_URL not set")
        auth = _build_auth_from_env()
        if auth is None:
            raise RuntimeError("No credentials available in env vars to fetch schema")
        fetch_schema_introspection(
            base_url=base_url,
            auth=auth,
            output_dir=str(schema_path.parent),
            experimental_apis=_env_experimental_apis(),
        )

    schema = _load_introspection(schema_path)
    cfg = _discover_config(schema)

    out_path = repo_root / "python" / "atlassian" / "graph" / "gen" / "jira_worklogs_api.py"
    out_path.parent.mkdir(parents=True, exist_ok=True)
    out_path.write_text(_render_python(cfg), encoding="utf-8")
    print(f"Wrote {out_path}")


if __name__ == "__main__":
    main()



import json

import httpx

from atlassian.auth import OAuthBearerAuth
from atlassian.graph.api.jira_issues import get_issue_by_key
from atlassian.graph.client import GraphQLClient


def test_jira_issue_by_key_graphql_mapping():
    def handler(request: httpx.Request) -> httpx.Response:
        payload = json.loads(request.content.decode("utf-8"))
        assert payload.get("operationName") == "JiraIssueByKey"
        assert payload.get("variables") == {"cloudId": "cloud-123", "key": "A-1"}
        return httpx.Response(
            200,
            json={
                "data": {
                    "issueByKey": {
                        "key": "A-1",
                        "issueType": {"name": "Bug"},
                        "status": {"name": "Done"},
                        "projectField": {"project": {"key": "A", "cloudId": "cloud-123"}},
                        "createdField": {"dateTime": "2021-01-01T00:00:00Z"},
                        "updatedField": {"dateTime": "2021-01-02T00:00:00Z"},
                        "resolutionDateField": {"dateTime": "2021-01-03T00:00:00Z"},
                        "assigneeField": {"user": {"accountId": "u1", "name": "User One"}},
                        "reporter": {"accountId": "u2", "name": "User Two"},
                    }
                }
            },
            request=request,
        )

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            http_client=http_client,
        )
        issue = get_issue_by_key(client, "cloud-123", "A-1")

    assert issue.key == "A-1"
    assert issue.project_key == "A"
    assert issue.issue_type == "Bug"
    assert issue.status == "Done"
    assert issue.assignee and issue.assignee.account_id == "u1"
    assert issue.reporter and issue.reporter.account_id == "u2"
    assert issue.resolved_at == "2021-01-03T00:00:00Z"



from __future__ import annotations

import json
import time
from datetime import datetime, timezone
from typing import Callable, Dict, List, Optional, Tuple

import httpx

from ..auth import AuthProvider
from ..errors import (
    GraphQLOperationError,
    LocalRateLimitError,
    RateLimitError,
    SerializationError,
    TransportError,
)
from ..logging import get_logger, sanitize_headers
from ..models import GraphQLErrorItem, GraphQLResult, parse_error_items
from ..retry import parse_retry_after
from .throttle import TokenBucket


class GraphQLClient:
    def __init__(
        self,
        base_url: str,
        auth: AuthProvider,
        timeout_seconds: float = 15.0,
        strict: bool = False,
        logger=None,
        user_agent: Optional[str] = None,
        max_retries_429: int = 2,
        max_wait_seconds: int = 60,
        enable_local_throttling: bool = False,
        sleeper: Callable[[float], None] | None = None,
        time_provider: Callable[[], datetime] | None = None,
        http_client: httpx.Client | None = None,
    ):
        if not base_url:
            raise ValueError("base_url is required")
        if auth is None:
            raise ValueError("auth is required")
        self.base_url = base_url.rstrip("/")
        self.auth = auth
        self.strict = strict
        self.max_retries_429 = max(0, max_retries_429)
        self.max_wait_seconds = max(0, max_wait_seconds)
        self.enable_local_throttling = enable_local_throttling
        self._logger = get_logger(logger)
        self._graphql_url = (
            self.base_url
            if self.base_url.endswith("/graphql")
            else f"{self.base_url}/graphql"
        )
        self._owns_client = http_client is None
        self._client = http_client if http_client is not None else httpx.Client(timeout=timeout_seconds)
        self._sleeper = sleeper if sleeper is not None else time.sleep
        self._now = (
            time_provider
            if time_provider is not None
            else lambda: datetime.now(timezone.utc)
        )
        self._bucket_capacity = 10000.0
        self._bucket_refill_rate = self._bucket_capacity / 60.0
        self._token_bucket = (
            TokenBucket(
                capacity=self._bucket_capacity,
                refill_rate_per_sec=self._bucket_refill_rate,
                now=self._now,
                sleeper=self._sleeper,
            )
            if self.enable_local_throttling
            else None
        )
        self._user_agent = user_agent or "atlassian-graphql-python/0.1.0"
        self._base_headers: List[tuple[str, str]] = [
            ("Content-Type", "application/json"),
            ("Accept", "application/json"),
            ("User-Agent", self._user_agent),
        ]

    def _consume_local_budget(self, estimated_cost: float) -> None:
        if self._token_bucket is None:
            return
        wait_time = self._token_bucket.consume(
            float(estimated_cost),
            float(self.max_wait_seconds),
        )
        self._logger.debug(
            "Local throttling applied",
            extra={
                "endpoint": self._graphql_url,
                "estimated_cost": estimated_cost,
                "wait_seconds": round(wait_time, 4),
            },
        )

    def _build_headers(self, experimental_apis: Optional[List[str]]) -> httpx.Headers:
        header_items: List[tuple[str, str]] = list(self._base_headers)
        if experimental_apis:
            for beta in experimental_apis:
                if beta:
                    header_items.append(("X-ExperimentalApi", beta))
        headers = httpx.Headers(header_items)
        self.auth.apply(headers)
        return headers

    def _extract_request_id(self, response: httpx.Response) -> Optional[str]:
        try:
            payload = response.json()
        except Exception:
            return None
        if isinstance(payload, dict):
            extensions = payload.get("extensions")
            if isinstance(extensions, dict):
                for key in ("requestId", "request_id", "requestid"):
                    value = extensions.get(key)
                    if isinstance(value, str):
                        return value
        return None

    def execute(
        self,
        query: str,
        variables: Optional[Dict] = None,
        operation_name: Optional[str] = None,
        experimental_apis: Optional[List[str]] = None,
        estimated_cost: int = 1,
    ) -> GraphQLResult:
        if not query or not query.strip():
            raise ValueError("query must be a non-empty string")

        payload: Dict[str, object] = {"query": query}
        if variables is not None:
            payload["variables"] = variables
        if operation_name:
            payload["operationName"] = operation_name

        cost_value = 1 if estimated_cost is None else estimated_cost
        if cost_value < 0:
            cost_value = 0

        retries = 0
        while True:
            attempt_number = retries + 1
            self._consume_local_budget(cost_value)
            headers = self._build_headers(experimental_apis)
            cookies = self.auth.get_cookies() if hasattr(self.auth, "get_cookies") else None
            start = time.perf_counter()
            try:
                response = self._client.post(
                    self._graphql_url,
                    headers=headers,
                    json=payload,
                    cookies=cookies,
                )
            except httpx.RequestError as exc:
                self._logger.error("HTTP request failed", exc_info=exc)
                raise TransportError(status_code=0, body_snippet=str(exc)) from exc

            try:
                duration = time.perf_counter() - start
                self._logger.debug(
                    "GraphQL request completed",
                    extra={
                        "operationName": operation_name,
                        "attempt": attempt_number,
                        "status_code": response.status_code,
                        "duration_sec": round(duration, 4),
                        "headers": sanitize_headers(headers),
                    },
                )

                if response.status_code == 429:
                    retry_header = response.headers.get("Retry-After")
                    request_id = self._extract_request_id(response)
                    try:
                        retry_at, parser_used = parse_retry_after(retry_header)
                        self._logger.debug(
                            "Parsed Retry-After header",
                            extra={
                                "retry_after": retry_header,
                                "parser": parser_used,
                                "retry_at": retry_at.isoformat(),
                                "operationName": operation_name,
                            },
                        )
                    except ValueError as exc:
                        self._logger.debug(
                            "Retry-After parsing failed",
                            extra={
                                "retry_after": retry_header,
                                "parser": None,
                                "operationName": operation_name,
                            },
                        )
                        raise RateLimitError(
                            retry_after=None,
                            attempts=attempt_number,
                            header_value=retry_header,
                        ) from exc

                    computed_wait = (retry_at - self._now()).total_seconds()
                    wait_seconds = computed_wait
                    if wait_seconds <= 0:
                        wait_seconds = 0.0

                    retry_allowed = retries < self.max_retries_429
                    over_cap = computed_wait > self.max_wait_seconds
                    self._logger.warning(
                        "Rate limited on GraphQL request",
                        extra={
                            "endpoint": self._graphql_url,
                            "operationName": operation_name,
                            "attempt": attempt_number,
                            "retry_at": retry_at.isoformat(),
                            "computed_wait_seconds": round(computed_wait, 4),
                            "wait_seconds": round(wait_seconds, 4),
                            "request_id": request_id,
                            "retrying": retry_allowed and not over_cap,
                        },
                    )
                    if over_cap:
                        raise RateLimitError(
                            retry_after=retry_at,
                            attempts=attempt_number,
                            header_value=retry_header,
                            wait_seconds=computed_wait,
                            max_wait_seconds=self.max_wait_seconds,
                        )
                    if not retry_allowed:
                        raise RateLimitError(
                            retry_after=retry_at,
                            attempts=attempt_number,
                            header_value=retry_header,
                            wait_seconds=computed_wait,
                        )
                    if wait_seconds > 0:
                        self._sleeper(wait_seconds)
                    retries += 1
                    continue

                if response.status_code >= 500:
                    raise TransportError(
                        status_code=response.status_code,
                        body_snippet=response.text[:200],
                    )
                if response.status_code >= 400:
                    raise TransportError(
                        status_code=response.status_code,
                        body_snippet=response.text[:200],
                    )

                try:
                    body = response.json()
                except json.JSONDecodeError as exc:
                    raise SerializationError(f"Failed to parse JSON: {exc}") from exc

                data = body.get("data") if isinstance(body, dict) else None
                errors = (
                    parse_error_items(body.get("errors"))
                    if isinstance(body, dict)
                    else None
                )
                extensions = body.get("extensions") if isinstance(body, dict) else None

                if self.strict and errors:
                    raise GraphQLOperationError(errors=errors, partial_data=data)

                return GraphQLResult(data=data, errors=errors, extensions=extensions)
            finally:
                response.close()

    def close(self) -> None:
        if self._owns_client:
            self._client.close()

    def __enter__(self) -> "GraphQLClient":
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()



# Code generated by python/tools/generate_jira_rest_models.py. DO NOT EDIT.
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from atlassian.errors import SerializationError

def _expect_dict(obj: Any, path: str) -> Dict[str, Any]:
    if not isinstance(obj, dict):
        raise SerializationError(f"Expected object at {path}")
    return obj


def _expect_list(obj: Any, path: str) -> List[Any]:
    if not isinstance(obj, list):
        raise SerializationError(f"Expected list at {path}")
    return obj


def _expect_str(obj: Any, path: str) -> str:
    if not isinstance(obj, str):
        raise SerializationError(f"Expected string at {path}")
    return obj


def _expect_bool(obj: Any, path: str) -> bool:
    if not isinstance(obj, bool):
        raise SerializationError(f"Expected boolean at {path}")
    return obj


def _expect_int(obj: Any, path: str) -> int:
    if isinstance(obj, bool) or not isinstance(obj, int):
        raise SerializationError(f"Expected integer at {path}")
    return obj


def _expect_obj(obj: Any, path: str) -> Dict[str, Any]:
    # Jira issue fields are modeled as a free-form object in the OpenAPI spec.
    return _expect_dict(obj, path)

@dataclass(frozen=True)
class UserDetails:
    account_id: Optional[str]
    display_name: Optional[str]
    email_address: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any, path: str) -> "UserDetails":
        raw = _expect_dict(obj, path)
        account_id: Optional[str] = None
        if raw.get("accountId") is not None:
            account_id = _expect_str(raw.get("accountId"), f"{path}.accountId")
        display_name: Optional[str] = None
        if raw.get("displayName") is not None:
            display_name = _expect_str(raw.get("displayName"), f"{path}.displayName")
        email_address: Optional[str] = None
        if raw.get("emailAddress") is not None:
            email_address = _expect_str(raw.get("emailAddress"), f"{path}.emailAddress")
        return UserDetails(
            account_id=account_id,
            display_name=display_name,
            email_address=email_address,
        )


@dataclass(frozen=True)
class Project:
    id: Optional[str]
    key: Optional[str]
    name: Optional[str]
    project_type_key: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any, path: str) -> "Project":
        raw = _expect_dict(obj, path)
        project_id: Optional[str] = None
        if raw.get("id") is not None:
            project_id = _expect_str(raw.get("id"), f"{path}.id")
        key: Optional[str] = None
        if raw.get("key") is not None:
            key = _expect_str(raw.get("key"), f"{path}.key")
        name: Optional[str] = None
        if raw.get("name") is not None:
            name = _expect_str(raw.get("name"), f"{path}.name")
        project_type_key: Optional[str] = None
        if raw.get("projectTypeKey") is not None:
            project_type_key = _expect_str(raw.get("projectTypeKey"), f"{path}.projectTypeKey")
        return Project(
            id=project_id,
            key=key,
            name=name,
            project_type_key=project_type_key,
        )


@dataclass(frozen=True)
class PageBeanProject:
    start_at: Optional[int]
    max_results: Optional[int]
    total: Optional[int]
    is_last: Optional[bool]
    values: List[Project]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "PageBeanProject":
        raw = _expect_dict(obj, path)
        start_at: Optional[int] = None
        if raw.get("startAt") is not None:
            start_at = _expect_int(raw.get("startAt"), f"{path}.startAt")
        max_results: Optional[int] = None
        if raw.get("maxResults") is not None:
            max_results = _expect_int(raw.get("maxResults"), f"{path}.maxResults")
        total: Optional[int] = None
        if raw.get("total") is not None:
            total = _expect_int(raw.get("total"), f"{path}.total")
        is_last: Optional[bool] = None
        if raw.get("isLast") is not None:
            is_last = _expect_bool(raw.get("isLast"), f"{path}.isLast")
        values_raw = raw.get("values")
        values_list = _expect_list(values_raw, f"{path}.values") if values_raw is not None else []
        values = [
            Project.from_dict(item, f"{path}.values[{idx}]")
            for idx, item in enumerate(values_list)
        ]
        return PageBeanProject(
            start_at=start_at,
            max_results=max_results,
            total=total,
            is_last=is_last,
            values=values,
        )


@dataclass(frozen=True)
class IssueBean:
    id: Optional[str]
    key: Optional[str]
    fields: Dict[str, Any]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "IssueBean":
        raw = _expect_dict(obj, path)
        issue_id: Optional[str] = None
        if raw.get("id") is not None:
            issue_id = _expect_str(raw.get("id"), f"{path}.id")
        key: Optional[str] = None
        if raw.get("key") is not None:
            key = _expect_str(raw.get("key"), f"{path}.key")
        fields_raw = raw.get("fields")
        fields = _expect_obj(fields_raw, f"{path}.fields") if fields_raw is not None else {}
        return IssueBean(id=issue_id, key=key, fields=fields)


@dataclass(frozen=True)
class SearchResults:
    start_at: Optional[int]
    max_results: Optional[int]
    total: Optional[int]
    issues: List[IssueBean]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "SearchResults":
        raw = _expect_dict(obj, path)
        start_at: Optional[int] = None
        if raw.get("startAt") is not None:
            start_at = _expect_int(raw.get("startAt"), f"{path}.startAt")
        max_results: Optional[int] = None
        if raw.get("maxResults") is not None:
            max_results = _expect_int(raw.get("maxResults"), f"{path}.maxResults")
        total: Optional[int] = None
        if raw.get("total") is not None:
            total = _expect_int(raw.get("total"), f"{path}.total")
        issues_raw = raw.get("issues")
        issues_list = _expect_list(issues_raw, f"{path}.issues") if issues_raw is not None else []
        issues = [
            IssueBean.from_dict(item, f"{path}.issues[{idx}]")
            for idx, item in enumerate(issues_list)
        ]
        return SearchResults(
            start_at=start_at,
            max_results=max_results,
            total=total,
            issues=issues,
        )


@dataclass(frozen=True)
class ChangeDetails:
    field: Optional[str]
    from_value: Optional[str] = None
    to_value: Optional[str] = None
    from_string: Optional[str] = None
    to_string: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any, path: str) -> "ChangeDetails":
        raw = _expect_dict(obj, path)
        field: Optional[str] = None
        if raw.get("field") is not None:
            field = _expect_str(raw.get("field"), f"{path}.field")
        from_value: Optional[str] = None
        if raw.get("from") is not None:
            from_value = _expect_str(raw.get("from"), f"{path}.from")
        to_value: Optional[str] = None
        if raw.get("to") is not None:
            to_value = _expect_str(raw.get("to"), f"{path}.to")
        from_string: Optional[str] = None
        if raw.get("fromString") is not None:
            from_string = _expect_str(raw.get("fromString"), f"{path}.fromString")
        to_string: Optional[str] = None
        if raw.get("toString") is not None:
            to_string = _expect_str(raw.get("toString"), f"{path}.toString")
        return ChangeDetails(
            field=field,
            from_value=from_value,
            to_value=to_value,
            from_string=from_string,
            to_string=to_string,
        )


@dataclass(frozen=True)
class Changelog:
    id: Optional[str]
    created: Optional[str]
    items: List[ChangeDetails]
    author: Optional[UserDetails] = None

    @staticmethod
    def from_dict(obj: Any, path: str) -> "Changelog":
        raw = _expect_dict(obj, path)
        event_id: Optional[str] = None
        if raw.get("id") is not None:
            event_id = _expect_str(raw.get("id"), f"{path}.id")
        created: Optional[str] = None
        if raw.get("created") is not None:
            created = _expect_str(raw.get("created"), f"{path}.created")
        author: Optional[UserDetails] = None
        if raw.get("author") is not None:
            author = UserDetails.from_dict(raw.get("author"), f"{path}.author")
        items_raw = raw.get("items")
        items_list = _expect_list(items_raw, f"{path}.items") if items_raw is not None else []
        items = [
            ChangeDetails.from_dict(item, f"{path}.items[{idx}]")
            for idx, item in enumerate(items_list)
        ]
        return Changelog(
            id=event_id,
            created=created,
            items=items,
            author=author,
        )


@dataclass(frozen=True)
class PageBeanChangelog:
    start_at: Optional[int]
    max_results: Optional[int]
    total: Optional[int]
    is_last: Optional[bool]
    values: List[Changelog]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "PageBeanChangelog":
        raw = _expect_dict(obj, path)
        start_at: Optional[int] = None
        if raw.get("startAt") is not None:
            start_at = _expect_int(raw.get("startAt"), f"{path}.startAt")
        max_results: Optional[int] = None
        if raw.get("maxResults") is not None:
            max_results = _expect_int(raw.get("maxResults"), f"{path}.maxResults")
        total: Optional[int] = None
        if raw.get("total") is not None:
            total = _expect_int(raw.get("total"), f"{path}.total")
        is_last: Optional[bool] = None
        if raw.get("isLast") is not None:
            is_last = _expect_bool(raw.get("isLast"), f"{path}.isLast")
        values_raw = raw.get("values")
        values_list = _expect_list(values_raw, f"{path}.values") if values_raw is not None else []
        values = [
            Changelog.from_dict(item, f"{path}.values[{idx}]")
            for idx, item in enumerate(values_list)
        ]
        return PageBeanChangelog(
            start_at=start_at,
            max_results=max_results,
            total=total,
            is_last=is_last,
            values=values,
        )


@dataclass(frozen=True)
class Worklog:
    id: Optional[str]
    started: Optional[str]
    time_spent_seconds: Optional[int]
    created: Optional[str]
    updated: Optional[str]
    author: Optional[UserDetails] = None

    @staticmethod
    def from_dict(obj: Any, path: str) -> "Worklog":
        raw = _expect_dict(obj, path)
        worklog_id: Optional[str] = None
        if raw.get("id") is not None:
            worklog_id = _expect_str(raw.get("id"), f"{path}.id")
        started: Optional[str] = None
        if raw.get("started") is not None:
            started = _expect_str(raw.get("started"), f"{path}.started")
        time_spent_seconds: Optional[int] = None
        if raw.get("timeSpentSeconds") is not None:
            time_spent_seconds = _expect_int(raw.get("timeSpentSeconds"), f"{path}.timeSpentSeconds")
        created: Optional[str] = None
        if raw.get("created") is not None:
            created = _expect_str(raw.get("created"), f"{path}.created")
        updated: Optional[str] = None
        if raw.get("updated") is not None:
            updated = _expect_str(raw.get("updated"), f"{path}.updated")
        author: Optional[UserDetails] = None
        if raw.get("author") is not None:
            author = UserDetails.from_dict(raw.get("author"), f"{path}.author")
        return Worklog(
            id=worklog_id,
            started=started,
            time_spent_seconds=time_spent_seconds,
            created=created,
            updated=updated,
            author=author,
        )


@dataclass(frozen=True)
class PageOfWorklogs:
    start_at: Optional[int]
    max_results: Optional[int]
    total: Optional[int]
    worklogs: List[Worklog]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "PageOfWorklogs":
        raw = _expect_dict(obj, path)
        start_at: Optional[int] = None
        if raw.get("startAt") is not None:
            start_at = _expect_int(raw.get("startAt"), f"{path}.startAt")
        max_results: Optional[int] = None
        if raw.get("maxResults") is not None:
            max_results = _expect_int(raw.get("maxResults"), f"{path}.maxResults")
        total: Optional[int] = None
        if raw.get("total") is not None:
            total = _expect_int(raw.get("total"), f"{path}.total")
        worklogs_raw = raw.get("worklogs")
        worklogs_list = _expect_list(worklogs_raw, f"{path}.worklogs") if worklogs_raw is not None else []
        worklogs = [
            Worklog.from_dict(item, f"{path}.worklogs[{idx}]")
            for idx, item in enumerate(worklogs_list)
        ]
        return PageOfWorklogs(
            start_at=start_at,
            max_results=max_results,
            total=total,
            worklogs=worklogs,
        )


@dataclass(frozen=True)
class PageBeanVersion:
    start_at: Optional[int]
    max_results: Optional[int]
    total: Optional[int]
    is_last: Optional[bool]
    values: List[Version]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "PageBeanVersion":
        raw = _expect_dict(obj, path)
        start_at: Optional[int] = None
        if raw.get("startAt") is not None:
            start_at = _expect_int(raw.get("startAt"), f"{path}.startAt")
        max_results: Optional[int] = None
        if raw.get("maxResults") is not None:
            max_results = _expect_int(raw.get("maxResults"), f"{path}.maxResults")
        total: Optional[int] = None
        if raw.get("total") is not None:
            total = _expect_int(raw.get("total"), f"{path}.total")
        is_last: Optional[bool] = None
        if raw.get("isLast") is not None:
            is_last = _expect_bool(raw.get("isLast"), f"{path}.isLast")
        values_raw = raw.get("values")
        values_list = _expect_list(values_raw, f"{path}.values") if values_raw is not None else []
        values = [
            Version.from_dict(item, f"{path}.values[{idx}]")
            for idx, item in enumerate(values_list)
        ]
        return PageBeanVersion(
            start_at=start_at,
            max_results=max_results,
            total=total,
            is_last=is_last,
            values=values,
        )


@dataclass(frozen=True)
class Version:
    id: Optional[str]
    name: Optional[str]
    project_id: Optional[int] = None
    released: Optional[bool] = None
    release_date: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any, path: str) -> "Version":
        raw = _expect_dict(obj, path)
        version_id: Optional[str] = None
        if raw.get("id") is not None:
            version_id = _expect_str(raw.get("id"), f"{path}.id")
        name: Optional[str] = None
        if raw.get("name") is not None:
            name = _expect_str(raw.get("name"), f"{path}.name")
        project_id: Optional[int] = None
        if raw.get("projectId") is not None:
            project_id = _expect_int(raw.get("projectId"), f"{path}.projectId")
        released: Optional[bool] = None
        if raw.get("released") is not None:
            released = _expect_bool(raw.get("released"), f"{path}.released")
        release_date: Optional[str] = None
        if raw.get("releaseDate") is not None:
            # Versions can have date-only strings "2010-07-06"
            release_date = _expect_str(raw.get("releaseDate"), f"{path}.releaseDate")
        return Version(
            id=version_id,
            name=name,
            project_id=project_id,
            released=released,
            release_date=release_date,
        )




from __future__ import annotations

import json
import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple


def _add_project_to_syspath() -> None:
    project_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(project_root))


_add_project_to_syspath()

from atlassian.auth import (  # noqa: E402
    BasicApiTokenAuth,
    CookieAuth,
    OAuthBearerAuth,
)
from atlassian.oauth_3lo import OAuthRefreshTokenAuth  # noqa: E402
from atlassian.graph.schema_fetcher import fetch_schema_introspection  # noqa: E402


@dataclass(frozen=True)
class _Config:
    cloud_id_type: str
    projects_first_type: str
    projects_after_type: str
    ops_first_type: str
    ops_after_type: str
    pageinfo_has_end_cursor: bool
    projects_edge_has_cursor: bool
    ops_edge_has_cursor: bool
    project_has_id: bool
    refetch_strategy: str  # "node" or "jira"
    project_type_name: str
    node_id_arg_type: Optional[str]
    jira_project_field_name: Optional[str]
    jira_project_key_arg_name: Optional[str]
    jira_project_key_arg_type: Optional[str]


def _env_experimental_apis() -> List[str]:
    raw = os.getenv("ATLASSIAN_GQL_EXPERIMENTAL_APIS", "")
    return [part.strip() for part in raw.split(",") if part.strip()]


def _maybe_strip_quotes(raw: str) -> str:
    if len(raw) >= 2 and raw[0] == raw[-1] and raw[0] in {"'", '"'}:
        return raw[1:-1]
    return raw


def _load_env_file(path: Path) -> None:
    if not path.exists():
        return
    for line in path.read_text(encoding="utf-8").splitlines():
        stripped = line.strip()
        if not stripped or stripped.startswith("#"):
            continue
        if stripped.startswith("export "):
            stripped = stripped[len("export ") :].strip()
        if "=" not in stripped:
            continue
        key, _, value = stripped.partition("=")
        key = key.strip()
        if not key or key in os.environ:
            continue
        os.environ[key] = _maybe_strip_quotes(value.strip())


def _build_auth_from_env():
    token = os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN")
    refresh_token = os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
    client_id = os.getenv("ATLASSIAN_CLIENT_ID")
    client_secret = os.getenv("ATLASSIAN_CLIENT_SECRET")
    email = os.getenv("ATLASSIAN_EMAIL")
    api_token = os.getenv("ATLASSIAN_API_TOKEN")
    cookies_json = os.getenv("ATLASSIAN_COOKIES_JSON")

    if refresh_token and client_id and client_secret:
        return OAuthRefreshTokenAuth(
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=refresh_token,
        )
    if token:
        if client_secret and token.strip() == client_secret.strip():
            raise ValueError(
                "ATLASSIAN_OAUTH_ACCESS_TOKEN appears to be set to ATLASSIAN_CLIENT_SECRET; "
                "set an OAuth access token (not the client secret)."
            )
        return OAuthBearerAuth(lambda: token)
    if email and api_token:
        return BasicApiTokenAuth(email, api_token)
    if cookies_json:
        try:
            cookies = json.loads(cookies_json)
        except json.JSONDecodeError:
            return None
        if isinstance(cookies, dict) and all(
            isinstance(k, str) and isinstance(v, str) for k, v in cookies.items()
        ):
            return CookieAuth(cookies)
    return None


def _load_introspection(path: Path) -> Dict[str, Any]:
    raw = json.loads(path.read_text(encoding="utf-8"))
    if isinstance(raw, dict) and "data" in raw and isinstance(raw["data"], dict):
        data = raw["data"]
    else:
        data = raw
    if not isinstance(data, dict) or "__schema" not in data:
        raise RuntimeError("Introspection JSON missing data.__schema")
    schema = data["__schema"]
    if not isinstance(schema, dict):
        raise RuntimeError("Introspection JSON data.__schema is not an object")
    return schema


def _types_map(schema: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
    types = schema.get("types")
    if not isinstance(types, list):
        raise RuntimeError("Introspection JSON missing __schema.types[]")
    out: Dict[str, Dict[str, Any]] = {}
    for t in types:
        if not isinstance(t, dict):
            continue
        name = t.get("name")
        if isinstance(name, str) and name:
            out[name] = t
    return out


def _unwrap_named_type(type_ref: Dict[str, Any]) -> Tuple[Optional[str], Optional[str], Dict[str, Any]]:
    cur = type_ref
    for _ in range(16):
        if not isinstance(cur, dict):
            break
        kind = cur.get("kind")
        name = cur.get("name")
        if isinstance(name, str) and name:
            return name, kind if isinstance(kind, str) else None, cur
        nxt = cur.get("ofType")
        if nxt is None:
            break
        cur = nxt
    return None, None, {}


def _type_ref_to_gql(type_ref: Dict[str, Any]) -> str:
    kind = type_ref.get("kind")
    if kind == "NON_NULL":
        of_type = type_ref.get("ofType")
        if not isinstance(of_type, dict):
            raise RuntimeError("Invalid NON_NULL typeRef")
        return f"{_type_ref_to_gql(of_type)}!"
    if kind == "LIST":
        of_type = type_ref.get("ofType")
        if not isinstance(of_type, dict):
            raise RuntimeError("Invalid LIST typeRef")
        return f"[{_type_ref_to_gql(of_type)}]"
    name = type_ref.get("name")
    if not isinstance(name, str) or not name:
        raise RuntimeError("Invalid named typeRef")
    return name


def _field(type_def: Dict[str, Any], name: str) -> Optional[Dict[str, Any]]:
    fields = type_def.get("fields")
    if not isinstance(fields, list):
        return None
    for f in fields:
        if isinstance(f, dict) and f.get("name") == name:
            return f
    return None


def _input_field(type_def: Dict[str, Any], name: str) -> Optional[Dict[str, Any]]:
    fields = type_def.get("inputFields")
    if not isinstance(fields, list):
        return None
    for f in fields:
        if isinstance(f, dict) and f.get("name") == name:
            return f
    return None


def _arg(field_def: Dict[str, Any], name: str) -> Optional[Dict[str, Any]]:
    args = field_def.get("args")
    if not isinstance(args, list):
        return None
    for a in args:
        if isinstance(a, dict) and a.get("name") == name:
            return a
    return None


def _discover_config(schema: Dict[str, Any]) -> _Config:
    types = _types_map(schema)
    missing: List[str] = []

    query_type = schema.get("queryType")
    query_name = query_type.get("name") if isinstance(query_type, dict) else None
    if not isinstance(query_name, str) or not query_name:
        raise RuntimeError("Introspection JSON missing __schema.queryType.name")
    query_def = types.get(query_name)
    if not query_def:
        raise RuntimeError(f"Missing query type definition: {query_name}")

    jira_field = _field(query_def, "jira")
    if not jira_field:
        missing.append(f"type {query_name}.fields.jira")
        raise RuntimeError("Missing required fields:\n- " + "\n- ".join(missing))
    jira_type_name, _, _ = _unwrap_named_type(jira_field.get("type") or {})
    if not jira_type_name or jira_type_name not in types:
        raise RuntimeError("Failed to resolve type for field Query.jira")
    jira_def = types[jira_type_name]

    all_projects_field = _field(jira_def, "allJiraProjects")
    if not all_projects_field:
        missing.append(f"type {jira_type_name}.fields.allJiraProjects")
        raise RuntimeError("Missing required fields:\n- " + "\n- ".join(missing))

    for arg_name in ("cloudId", "filter", "first", "after"):
        if not _arg(all_projects_field, arg_name):
            missing.append(f"field jira.allJiraProjects.args.{arg_name}")

    cloud_id_type = ""
    projects_first_type = ""
    projects_after_type = ""

    cloud_arg = _arg(all_projects_field, "cloudId")
    if cloud_arg and isinstance(cloud_arg.get("type"), dict):
        cloud_id_type = _type_ref_to_gql(cloud_arg["type"])

    first_arg = _arg(all_projects_field, "first")
    if first_arg and isinstance(first_arg.get("type"), dict):
        projects_first_type = _type_ref_to_gql(first_arg["type"])

    after_arg = _arg(all_projects_field, "after")
    if after_arg and isinstance(after_arg.get("type"), dict):
        projects_after_type = _type_ref_to_gql(after_arg["type"])

    filter_arg = _arg(all_projects_field, "filter")
    if filter_arg and isinstance(filter_arg.get("type"), dict):
        filter_type_name, _, _ = _unwrap_named_type(filter_arg["type"])
        filter_def = types.get(filter_type_name or "")
        if not filter_def:
            missing.append("field jira.allJiraProjects.args.filter.type")
        else:
            types_field = _input_field(filter_def, "types")
            if not types_field:
                missing.append(f"type {filter_type_name}.inputFields.types")
            else:
                tref = types_field.get("type")
                if not isinstance(tref, dict):
                    missing.append(f"type {filter_type_name}.inputFields.types.type")

    conn_type_name, _, _ = _unwrap_named_type(all_projects_field.get("type") or {})
    conn_def = types.get(conn_type_name or "")
    if not conn_def:
        missing.append("field jira.allJiraProjects.type")
        raise RuntimeError("Missing required fields:\n- " + "\n- ".join(missing))

    page_info_field = _field(conn_def, "pageInfo")
    edges_field = _field(conn_def, "edges")
    if not page_info_field:
        missing.append(f"type {conn_type_name}.fields.pageInfo")
    if not edges_field:
        missing.append(f"type {conn_type_name}.fields.edges")
    if missing:
        raise RuntimeError("Missing required fields:\n- " + "\n- ".join(missing))

    pageinfo_type_name, _, _ = _unwrap_named_type(page_info_field.get("type") or {})
    pageinfo_def = types.get(pageinfo_type_name or "")
    if not pageinfo_def:
        raise RuntimeError(f"Missing PageInfo type definition: {pageinfo_type_name}")

    has_next = _field(pageinfo_def, "hasNextPage")
    if not has_next:
        raise RuntimeError(f"Missing PageInfo.hasNextPage on {pageinfo_type_name}")
    pageinfo_has_end_cursor = _field(pageinfo_def, "endCursor") is not None

    edges_type_name, _, _ = _unwrap_named_type(edges_field.get("type") or {})
    edges_def = types.get(edges_type_name or "")
    if not edges_def:
        raise RuntimeError(f"Missing edge type definition: {edges_type_name}")
    projects_edge_has_cursor = _field(edges_def, "cursor") is not None

    node_field = _field(edges_def, "node")
    if not node_field:
        raise RuntimeError(f"Missing edge.node on {edges_type_name}")
    project_type_name, _, _ = _unwrap_named_type(node_field.get("type") or {})
    project_def = types.get(project_type_name or "")
    if not project_def:
        raise RuntimeError(f"Missing project type definition: {project_type_name}")

    project_has_id = _field(project_def, "id") is not None
    if not _field(project_def, "key"):
        missing.append(f"type {project_type_name}.fields.key")
    if not _field(project_def, "name"):
        missing.append(f"type {project_type_name}.fields.name")
    ops_field = _field(project_def, "opsgenieTeamsAvailableToLinkWith")
    if not ops_field:
        missing.append(f"type {project_type_name}.fields.opsgenieTeamsAvailableToLinkWith")
    else:
        for arg_name in ("first", "after"):
            if not _arg(ops_field, arg_name):
                missing.append(f"field {project_type_name}.opsgenieTeamsAvailableToLinkWith.args.{arg_name}")

    if missing:
        raise RuntimeError("Missing required fields:\n- " + "\n- ".join(missing))

    ops_first_type = ""
    ops_after_type = ""
    ops_first_arg = _arg(ops_field, "first")
    if ops_first_arg and isinstance(ops_first_arg.get("type"), dict):
        ops_first_type = _type_ref_to_gql(ops_first_arg["type"])
    ops_after_arg = _arg(ops_field, "after")
    if ops_after_arg and isinstance(ops_after_arg.get("type"), dict):
        ops_after_type = _type_ref_to_gql(ops_after_arg["type"])

    ops_conn_type_name, _, _ = _unwrap_named_type(ops_field.get("type") or {})
    ops_conn_def = types.get(ops_conn_type_name or "")
    if not ops_conn_def:
        raise RuntimeError(f"Missing opsgenie connection type: {ops_conn_type_name}")

    ops_page_info_field = _field(ops_conn_def, "pageInfo")
    ops_edges_field = _field(ops_conn_def, "edges")
    if not ops_page_info_field:
        missing.append(f"type {ops_conn_type_name}.fields.pageInfo")
    if not ops_edges_field:
        missing.append(f"type {ops_conn_type_name}.fields.edges")
    if missing:
        raise RuntimeError("Missing required fields:\n- " + "\n- ".join(missing))

    ops_edges_type_name, _, _ = _unwrap_named_type(ops_edges_field.get("type") or {})
    ops_edges_def = types.get(ops_edges_type_name or "")
    if not ops_edges_def:
        raise RuntimeError(f"Missing opsgenie edge type: {ops_edges_type_name}")
    ops_edge_has_cursor = _field(ops_edges_def, "cursor") is not None

    ops_node_field = _field(ops_edges_def, "node")
    if not ops_node_field:
        raise RuntimeError(f"Missing opsgenie edge.node on {ops_edges_type_name}")
    ops_team_type_name, _, _ = _unwrap_named_type(ops_node_field.get("type") or {})
    ops_team_def = types.get(ops_team_type_name or "")
    if not ops_team_def:
        raise RuntimeError(f"Missing opsgenie team type: {ops_team_type_name}")
    if not _field(ops_team_def, "id"):
        missing.append(f"type {ops_team_type_name}.fields.id")
    if not _field(ops_team_def, "name"):
        missing.append(f"type {ops_team_type_name}.fields.name")
    if missing:
        raise RuntimeError("Missing required fields:\n- " + "\n- ".join(missing))

    refetch_strategy = "jira"
    node_id_arg_type: Optional[str] = None
    jira_project_field_name: Optional[str] = None
    jira_project_key_arg_name: Optional[str] = None
    jira_project_key_arg_type: Optional[str] = None

    node_field_def = _field(query_def, "node")
    if node_field_def:
        id_arg = _arg(node_field_def, "id")
        if id_arg and isinstance(id_arg.get("type"), dict) and project_has_id:
            node_id_arg_type = _type_ref_to_gql(id_arg["type"])
            refetch_strategy = "node"

    if refetch_strategy != "node":
        candidates: List[Tuple[str, Dict[str, Any]]] = []
        for f in jira_def.get("fields", []):
            if not isinstance(f, dict):
                continue
            f_type_name, _, _ = _unwrap_named_type(f.get("type") or {})
            if f_type_name == project_type_name:
                candidates.append((f.get("name") or "", f))

        candidates = sorted(candidates, key=lambda item: item[0])
        for field_name, f in candidates:
            if not field_name:
                continue
            if not _arg(f, "cloudId"):
                continue
            for key_arg_name in ("key", "projectKey"):
                key_arg = _arg(f, key_arg_name)
                if key_arg and isinstance(key_arg.get("type"), dict):
                    jira_project_field_name = field_name
                    jira_project_key_arg_name = key_arg_name
                    jira_project_key_arg_type = _type_ref_to_gql(key_arg["type"])
                    break
            if jira_project_field_name:
                break

        if not jira_project_field_name:
            raise RuntimeError(
                "Unable to determine a per-project refetch strategy for nested opsgenie pagination. "
                "Expected Query.node(id: ...) support or a jira.<project>(cloudId, key) field returning the Jira project type."
            )

    return _Config(
        cloud_id_type=cloud_id_type,
        projects_first_type=projects_first_type,
        projects_after_type=projects_after_type,
        ops_first_type=ops_first_type,
        ops_after_type=ops_after_type,
        pageinfo_has_end_cursor=pageinfo_has_end_cursor,
        projects_edge_has_cursor=projects_edge_has_cursor,
        ops_edge_has_cursor=ops_edge_has_cursor,
        project_has_id=project_has_id,
        refetch_strategy=refetch_strategy,
        project_type_name=project_type_name,
        node_id_arg_type=node_id_arg_type,
        jira_project_field_name=jira_project_field_name,
        jira_project_key_arg_name=jira_project_key_arg_name,
        jira_project_key_arg_type=jira_project_key_arg_type,
    )


def _render_python(cfg: _Config) -> str:
    pageinfo_select = "hasNextPage"
    if cfg.pageinfo_has_end_cursor:
        pageinfo_select += " endCursor"

    project_edge_select = "node {"
    if cfg.projects_edge_has_cursor:
        project_edge_select = "cursor\n      node {"

    ops_edge_select = "node {"
    if cfg.ops_edge_has_cursor:
        ops_edge_select = "cursor\n            node {"

    project_id_select = "id\n          " if cfg.project_has_id else ""

    projects_page_query_template = f"""\
query JiraProjectsPage(
  $cloudId: {cfg.cloud_id_type},
  $first: {cfg.projects_first_type},
  $after: {cfg.projects_after_type},
  $opsFirst: {cfg.ops_first_type}
) {{
  jira {{
    projects: allJiraProjects(
      cloudId: $cloudId,
      filter: {{ types: [__PROJECT_TYPES__] }},
      first: $first,
      after: $after
    ) {{
      pageInfo {{ {pageinfo_select} }}
      edges {{
        {project_edge_select}
          {project_id_select}key
          name
          opsgenieTeams: opsgenieTeamsAvailableToLinkWith(first: $opsFirst) {{
            pageInfo {{ {pageinfo_select} }}
            edges {{
              {ops_edge_select}
                id
                name
              }}
            }}
          }}
        }}
      }}
    }}
  }}
}}
"""

    if cfg.refetch_strategy == "node":
        if not cfg.node_id_arg_type:
            raise RuntimeError("invalid config: node strategy missing node_id_arg_type")
        ops_query = f"""\
query JiraProjectOpsgenieTeamsPage(
  $projectId: {cfg.node_id_arg_type},
  $first: {cfg.ops_first_type},
  $after: {cfg.ops_after_type}
) {{
  project: node(id: $projectId) {{
    ... on {cfg.project_type_name} {{
      opsgenieTeams: opsgenieTeamsAvailableToLinkWith(first: $first, after: $after) {{
        pageInfo {{ {pageinfo_select} }}
        edges {{
          {ops_edge_select}
            id
            name
          }}
        }}
      }}
    }}
  }}
}}
"""
    else:
        if not cfg.jira_project_field_name or not cfg.jira_project_key_arg_name or not cfg.jira_project_key_arg_type:
            raise RuntimeError("invalid config: jira strategy missing project lookup details")
        ops_query = f"""\
query JiraProjectOpsgenieTeamsPage(
  $cloudId: {cfg.cloud_id_type},
  $projectKey: {cfg.jira_project_key_arg_type},
  $first: {cfg.ops_first_type},
  $after: {cfg.ops_after_type}
) {{
  jira {{
    project: {cfg.jira_project_field_name}(cloudId: $cloudId, {cfg.jira_project_key_arg_name}: $projectKey) {{
      opsgenieTeams: opsgenieTeamsAvailableToLinkWith(first: $first, after: $after) {{
        pageInfo {{ {pageinfo_select} }}
        edges {{
          {ops_edge_select}
            id
            name
          }}
        }}
      }}
    }}
  }}
}}
"""

    return f"""\
# Code generated by python/tools/generate_jira_project_models.py. DO NOT EDIT.
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence

from atlassian.errors import SerializationError

PAGEINFO_HAS_END_CURSOR = {str(cfg.pageinfo_has_end_cursor)}
PROJECTS_EDGE_HAS_CURSOR = {str(cfg.projects_edge_has_cursor)}
OPSGENIE_EDGE_HAS_CURSOR = {str(cfg.ops_edge_has_cursor)}
PROJECT_HAS_ID = {str(cfg.project_has_id)}
REFETCH_STRATEGY = {cfg.refetch_strategy!r}

JIRA_PROJECTS_PAGE_QUERY_TEMPLATE = {projects_page_query_template!r}
JIRA_PROJECT_OPSGENIE_TEAMS_PAGE_QUERY = {ops_query!r}


def build_jira_projects_page_query(project_types: Sequence[str]) -> str:
    if not project_types:
        raise ValueError("project_types must be non-empty")
    cleaned: List[str] = []
    for raw in project_types:
        if not isinstance(raw, str):
            raise ValueError("project_types must be strings")
        value = raw.strip().upper()
        if not value or not value.replace("_", "").isalnum() or not value[0].isalpha():
            raise ValueError(f"invalid Jira project type enum token: {{raw!r}}")
        cleaned.append(value)
    joined = ", ".join(cleaned)
    return JIRA_PROJECTS_PAGE_QUERY_TEMPLATE.replace("__PROJECT_TYPES__", joined)


def _expect_dict(obj: Any, path: str) -> Dict[str, Any]:
    if not isinstance(obj, dict):
        raise SerializationError(f"Expected object at {{path}}")
    return obj


def _expect_list(obj: Any, path: str) -> List[Any]:
    if not isinstance(obj, list):
        raise SerializationError(f"Expected list at {{path}}")
    return obj


def _expect_str(obj: Any, path: str) -> str:
    if not isinstance(obj, str):
        raise SerializationError(f"Expected string at {{path}}")
    return obj


def _expect_bool(obj: Any, path: str) -> bool:
    if not isinstance(obj, bool):
        raise SerializationError(f"Expected boolean at {{path}}")
    return obj


@dataclass(frozen=True)
class PageInfo:
    has_next_page: bool
    end_cursor: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any, path: str) -> "PageInfo":
        raw = _expect_dict(obj, path)
        has_next = _expect_bool(raw.get("hasNextPage"), f"{{path}}.hasNextPage")
        end_cursor: Optional[str] = None
        if PAGEINFO_HAS_END_CURSOR:
            value = raw.get("endCursor")
            if value is not None:
                end_cursor = _expect_str(value, f"{{path}}.endCursor")
        return PageInfo(has_next_page=has_next, end_cursor=end_cursor)


@dataclass(frozen=True)
class OpsgenieTeamNode:
    id: str
    name: str

    @staticmethod
    def from_dict(obj: Any, path: str) -> "OpsgenieTeamNode":
        raw = _expect_dict(obj, path)
        return OpsgenieTeamNode(
            id=_expect_str(raw.get("id"), f"{{path}}.id"),
            name=_expect_str(raw.get("name"), f"{{path}}.name"),
        )


@dataclass(frozen=True)
class OpsgenieTeamEdge:
    cursor: Optional[str]
    node: OpsgenieTeamNode

    @staticmethod
    def from_dict(obj: Any, path: str) -> "OpsgenieTeamEdge":
        raw = _expect_dict(obj, path)
        cursor: Optional[str] = None
        if OPSGENIE_EDGE_HAS_CURSOR:
            value = raw.get("cursor")
            if value is not None:
                cursor = _expect_str(value, f"{{path}}.cursor")
        node = OpsgenieTeamNode.from_dict(raw.get("node"), f"{{path}}.node")
        return OpsgenieTeamEdge(cursor=cursor, node=node)


@dataclass(frozen=True)
class OpsgenieTeamsConnection:
    page_info: PageInfo
    edges: List[OpsgenieTeamEdge]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "OpsgenieTeamsConnection":
        raw = _expect_dict(obj, path)
        page_info = PageInfo.from_dict(raw.get("pageInfo"), f"{{path}}.pageInfo")
        edges_list = _expect_list(raw.get("edges"), f"{{path}}.edges")
        edges = [
            OpsgenieTeamEdge.from_dict(item, f"{{path}}.edges[{{idx}}]")
            for idx, item in enumerate(edges_list)
        ]
        return OpsgenieTeamsConnection(page_info=page_info, edges=edges)


@dataclass(frozen=True)
class JiraProjectNode:
    id: Optional[str]
    key: str
    name: str
    opsgenie_teams: OpsgenieTeamsConnection

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraProjectNode":
        raw = _expect_dict(obj, path)
        project_id: Optional[str] = None
        if PROJECT_HAS_ID:
            value = raw.get("id")
            if value is not None:
                project_id = _expect_str(value, f"{{path}}.id")
        return JiraProjectNode(
            id=project_id,
            key=_expect_str(raw.get("key"), f"{{path}}.key"),
            name=_expect_str(raw.get("name"), f"{{path}}.name"),
            opsgenie_teams=OpsgenieTeamsConnection.from_dict(
                raw.get("opsgenieTeams"), f"{{path}}.opsgenieTeams"
            ),
        )


@dataclass(frozen=True)
class JiraProjectEdge:
    cursor: Optional[str]
    node: JiraProjectNode

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraProjectEdge":
        raw = _expect_dict(obj, path)
        cursor: Optional[str] = None
        if PROJECTS_EDGE_HAS_CURSOR:
            value = raw.get("cursor")
            if value is not None:
                cursor = _expect_str(value, f"{{path}}.cursor")
        node = JiraProjectNode.from_dict(raw.get("node"), f"{{path}}.node")
        return JiraProjectEdge(cursor=cursor, node=node)


@dataclass(frozen=True)
class JiraProjectsConnection:
    page_info: PageInfo
    edges: List[JiraProjectEdge]

    @staticmethod
    def from_dict(obj: Any, path: str) -> "JiraProjectsConnection":
        raw = _expect_dict(obj, path)
        page_info = PageInfo.from_dict(raw.get("pageInfo"), f"{{path}}.pageInfo")
        edges_list = _expect_list(raw.get("edges"), f"{{path}}.edges")
        edges = [
            JiraProjectEdge.from_dict(item, f"{{path}}.edges[{{idx}}]")
            for idx, item in enumerate(edges_list)
        ]
        return JiraProjectsConnection(page_info=page_info, edges=edges)


@dataclass(frozen=True)
class JiraProjectsPageData:
    projects: JiraProjectsConnection

    @staticmethod
    def from_dict(obj: Any, path: str = "data.jira") -> "JiraProjectsPageData":
        raw = _expect_dict(obj, path)
        return JiraProjectsPageData(
            projects=JiraProjectsConnection.from_dict(raw.get("projects"), f"{{path}}.projects"),
        )


def parse_jira_projects_page(data: Any) -> JiraProjectsPageData:
    root = _expect_dict(data, "data")
    jira = root.get("jira")
    if jira is None:
        raise SerializationError("Missing data.jira")
    return JiraProjectsPageData.from_dict(jira, "data.jira")


def parse_project_opsgenie_teams(data: Any) -> OpsgenieTeamsConnection:
    root = _expect_dict(data, "data")
    if REFETCH_STRATEGY == "node":
        project = root.get("project")
        if project is None:
            raise SerializationError("Missing data.project")
        project_obj = _expect_dict(project, "data.project")
        inner = project_obj.get("opsgenieTeams")
        if inner is None:
            raise SerializationError("Missing data.project.opsgenieTeams")
        return OpsgenieTeamsConnection.from_dict(inner, "data.project.opsgenieTeams")

    jira = root.get("jira")
    if jira is None:
        raise SerializationError("Missing data.jira")
    jira_obj = _expect_dict(jira, "data.jira")
    project = jira_obj.get("project")
    if project is None:
        raise SerializationError("Missing data.jira.project")
    project_obj = _expect_dict(project, "data.jira.project")
    inner = project_obj.get("opsgenieTeams")
    if inner is None:
        raise SerializationError("Missing data.jira.project.opsgenieTeams")
    return OpsgenieTeamsConnection.from_dict(inner, "data.jira.project.opsgenieTeams")
"""


def main(argv: Sequence[str]) -> int:
    repo_root = Path(__file__).resolve().parents[2]
    token_file = os.getenv("ATLASSIAN_OAUTH_TOKEN_FILE")
    if not token_file:
        token_file = str(repo_root / "oauth_tokens.txt")
    _load_env_file(Path(token_file))

    schema_path = repo_root / "graphql" / "schema.introspection.json"

    if not schema_path.exists():
        base_url = os.getenv("ATLASSIAN_GQL_BASE_URL")
        if not base_url and (
            os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN")
            or os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
        ):
            base_url = "https://api.atlassian.com"
        try:
            auth = _build_auth_from_env()
        except ValueError as exc:
            print(str(exc), file=sys.stderr)
            return 2
        if not base_url or auth is None:
            print(
                f"Missing {schema_path}. Set ATLASSIAN_GQL_BASE_URL (required for non-OAuth auth modes) and credentials, "
                "or run `make graphql-schema` first.",
                file=sys.stderr,
            )
            return 2
        fetch_schema_introspection(
            base_url,
            auth,
            output_dir=schema_path.parent,
            experimental_apis=_env_experimental_apis(),
        )

    schema = _load_introspection(schema_path)
    cfg = _discover_config(schema)
    output_py = repo_root / "python" / "atlassian" / "graph" / "gen" / "jira_projects_api.py"
    output_py.parent.mkdir(parents=True, exist_ok=True)
    output_py.write_text(_render_python(cfg), encoding="utf-8")

    print(f"Wrote {output_py}")
    return 0


if __name__ == "__main__":
    raise SystemExit(main(sys.argv[1:]))



import pytest

from atlassian.rest.gen.jira_agile_api import Sprint
from atlassian.rest.mappers.jira_sprints import map_sprint


def test_sprint_mapper_trims_fields():
    sprint = Sprint(
        id=100,
        name="  Sprint 1  ",
        state="  active  ",
        start_date="  2021-01-01T00:00:00.000Z  ",
        end_date="  2021-01-15T00:00:00.000Z  ",
        complete_date=None,
        origin_board_id=10,
        goal="  Goal text  ",
    )
    mapped = map_sprint(sprint=sprint)
    assert mapped.id == "100"
    assert mapped.name == "Sprint 1"
    assert mapped.state == "active"
    assert mapped.start_at == "2021-01-01T00:00:00.000Z"
    assert mapped.end_at == "2021-01-15T00:00:00.000Z"
    assert mapped.complete_at is None


def test_sprint_mapper_handles_optional_dates():
    sprint = Sprint(
        id=101,
        name="Sprint 2",
        state="closed",
        start_date=None,
        end_date=None,
        complete_date="2021-01-30T12:00:00.000Z",
        origin_board_id=None,
        goal=None,
    )
    mapped = map_sprint(sprint=sprint)
    assert mapped.id == "101"
    assert mapped.name == "Sprint 2"
    assert mapped.state == "closed"
    assert mapped.start_at is None
    assert mapped.end_at is None
    assert mapped.complete_at == "2021-01-30T12:00:00.000Z"


def test_sprint_mapper_requires_id():
    sprint = Sprint(
        id=None,
        name="Sprint",
        state="active",
        start_date=None,
        end_date=None,
        complete_date=None,
        origin_board_id=None,
        goal=None,
    )
    with pytest.raises(ValueError, match="sprint.id is required"):
        map_sprint(sprint=sprint)


def test_sprint_mapper_requires_name():
    sprint = Sprint(
        id=100,
        name="",
        state="active",
        start_date=None,
        end_date=None,
        complete_date=None,
        origin_board_id=None,
        goal=None,
    )
    with pytest.raises(ValueError, match="sprint.name is required"):
        map_sprint(sprint=sprint)


def test_sprint_mapper_requires_name_not_whitespace():
    sprint = Sprint(
        id=100,
        name="   ",
        state="active",
        start_date=None,
        end_date=None,
        complete_date=None,
        origin_board_id=None,
        goal=None,
    )
    with pytest.raises(ValueError, match="sprint.name is required"):
        map_sprint(sprint=sprint)


def test_sprint_mapper_requires_state():
    sprint = Sprint(
        id=100,
        name="Sprint",
        state="",
        start_date=None,
        end_date=None,
        complete_date=None,
        origin_board_id=None,
        goal=None,
    )
    with pytest.raises(ValueError, match="sprint.state is required"):
        map_sprint(sprint=sprint)


def test_sprint_mapper_requires_sprint():
    with pytest.raises(ValueError, match="sprint is required"):
        map_sprint(sprint=None)  # type: ignore[arg-type]



from __future__ import annotations

from typing import Optional

from ...canonical_models import JiraSprint
from ..gen import jira_sprints_api as api


def _require_non_empty(value: Optional[str], path: str) -> str:
    if not isinstance(value, str) or not value.strip():
        raise ValueError(f"{path} is required")
    return value.strip()


def map_sprint(*, sprint: api.JiraSprintNode) -> JiraSprint:
    if sprint is None:
        raise ValueError("sprint is required")

    sprint_id = _require_non_empty(sprint.sprint_id, "sprint.sprintId")
    name = _require_non_empty(sprint.name, "sprint.name")
    state = _require_non_empty(sprint.state, "sprint.state")

    start_at = sprint.start_date.strip() if isinstance(sprint.start_date, str) and sprint.start_date.strip() else None
    end_at = sprint.end_date.strip() if isinstance(sprint.end_date, str) and sprint.end_date.strip() else None
    complete_at = (
        sprint.completion_date.strip()
        if isinstance(sprint.completion_date, str) and sprint.completion_date.strip()
        else None
    )

    return JiraSprint(
        id=sprint_id,
        name=name,
        state=state,
        start_at=start_at,
        end_at=end_at,
        complete_at=complete_at,
    )



from __future__ import annotations

import json
import os
from typing import Iterator, List, Optional, Sequence

from ...auth import BasicApiTokenAuth, CookieAuth, OAuthBearerAuth
from ...canonical_models import CanonicalProjectWithOpsgenieTeams
from ...errors import GraphQLOperationError, SerializationError
from ...oauth_3lo import OAuthRefreshTokenAuth
from ..client import GraphQLClient
from ..gen import jira_projects_api as api
from ..mappers.jira_projects import map_project_with_opsgenie_teams


def _normalize_project_types(project_types: Sequence[str]) -> List[str]:
    cleaned: List[str] = []
    for raw in project_types:
        if not isinstance(raw, str):
            raise ValueError("project_types must be strings")
        value = raw.strip().upper()
        if not value:
            continue
        cleaned.append(value)
    if not cleaned:
        raise ValueError("project_types must be non-empty")
    return cleaned


def _env_experimental_apis() -> List[str]:
    raw = os.getenv("ATLASSIAN_GQL_EXPERIMENTAL_APIS", "")
    return [part.strip() for part in raw.split(",") if part.strip()]


def _auth_from_env():
    token = os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN")
    refresh_token = os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
    client_id = os.getenv("ATLASSIAN_CLIENT_ID")
    client_secret = os.getenv("ATLASSIAN_CLIENT_SECRET")
    email = os.getenv("ATLASSIAN_EMAIL")
    api_token = os.getenv("ATLASSIAN_API_TOKEN")
    cookies_json = os.getenv("ATLASSIAN_COOKIES_JSON")

    if refresh_token and client_id and client_secret:
        return OAuthRefreshTokenAuth(
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=refresh_token,
        )
    if token:
        if client_secret and token.strip() == client_secret.strip():
            raise ValueError(
                "ATLASSIAN_OAUTH_ACCESS_TOKEN appears to be set to ATLASSIAN_CLIENT_SECRET; "
                "set an OAuth access token (not the client secret)."
            )
        return OAuthBearerAuth(lambda: token)
    if email and api_token:
        return BasicApiTokenAuth(email, api_token)
    if cookies_json:
        try:
            cookies = json.loads(cookies_json)
        except json.JSONDecodeError:
            return None
        if isinstance(cookies, dict) and all(
            isinstance(k, str) and isinstance(v, str) for k, v in cookies.items()
        ):
            return CookieAuth(cookies)
    return None


def _next_after_from_pageinfo(
    *,
    has_next_page: bool,
    end_cursor: Optional[str],
    edge_has_cursor: bool,
    edges_cursors: Sequence[Optional[str]],
    path: str,
) -> Optional[str]:
    if not has_next_page:
        return None
    if api.PAGEINFO_HAS_END_CURSOR and end_cursor:
        return end_cursor
    if edge_has_cursor:
        for cursor in reversed(edges_cursors):
            if cursor:
                return cursor
    raise SerializationError(f"Pagination cursor missing for {path}")


def iter_projects_with_opsgenie_linkable_teams(
    client: GraphQLClient,
    cloud_id: str,
    project_types: List[str],
    page_size: int = 50,
    *,
    experimental_apis: Optional[Sequence[str]] = None,
) -> Iterator[CanonicalProjectWithOpsgenieTeams]:
    cloud_id_clean = (cloud_id or "").strip()
    if not cloud_id_clean:
        raise ValueError("cloud_id is required")
    if page_size <= 0:
        raise ValueError("page_size must be > 0")

    normalized_types = _normalize_project_types(project_types)
    query = api.build_jira_projects_page_query(normalized_types)

    after: Optional[str] = None
    seen_after: set[str] = set()

    while True:
        variables = {
            "cloudId": cloud_id_clean,
            "first": page_size,
            "after": after,
            "opsFirst": page_size,
        }
        result = client.execute(
            query,
            variables=variables,
            operation_name="JiraProjectsPage",
            experimental_apis=list(experimental_apis) if experimental_apis else None,
        )
        if result.data is None:
            raise SerializationError("Missing GraphQL data in response")

        try:
            page = api.parse_jira_projects_page(result.data)
        except SerializationError as exc:
            if result.errors:
                raise GraphQLOperationError(errors=result.errors, partial_data=result.data) from exc
            raise
        conn = page.projects

        for edge in conn.edges:
            project = edge.node
            teams_conn = project.opsgenie_teams
            teams = [e.node for e in teams_conn.edges]
            if teams_conn.page_info.has_next_page:
                teams.extend(
                    _paginate_opsgenie_teams(
                        client=client,
                        cloud_id=cloud_id_clean,
                        project=project,
                        initial_connection=teams_conn,
                        page_size=page_size,
                        experimental_apis=experimental_apis,
                    )
                )

            yield map_project_with_opsgenie_teams(
                cloud_id=cloud_id_clean,
                project=project,
                opsgenie_teams=teams,
            )

        next_after = _next_after_from_pageinfo(
            has_next_page=conn.page_info.has_next_page,
            end_cursor=conn.page_info.end_cursor,
            edge_has_cursor=api.PROJECTS_EDGE_HAS_CURSOR,
            edges_cursors=[e.cursor for e in conn.edges],
            path="jira.projects",
        )
        if next_after is None:
            break
        if next_after in seen_after:
            raise SerializationError("Pagination cursor repeated; aborting to prevent infinite loop")
        seen_after.add(next_after)
        after = next_after


def list_projects_with_opsgenie_linkable_teams(
    cloud_id: str,
    project_types: List[str],
    page_size: int = 50,
) -> Iterator[CanonicalProjectWithOpsgenieTeams]:
    base_url = os.getenv("ATLASSIAN_GQL_BASE_URL")
    auth = _auth_from_env()
    if not base_url and (
        os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN")
        or os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
    ):
        base_url = "https://api.atlassian.com"
    if not base_url or auth is None:
        raise ValueError(
            "Missing ATLASSIAN_GQL_BASE_URL and/or credentials. "
            "Set ATLASSIAN_OAUTH_ACCESS_TOKEN, or ATLASSIAN_OAUTH_REFRESH_TOKEN + (ATLASSIAN_CLIENT_ID + ATLASSIAN_CLIENT_SECRET), "
            "or (ATLASSIAN_EMAIL + ATLASSIAN_API_TOKEN), or ATLASSIAN_COOKIES_JSON."
        )

    experimental_apis = _env_experimental_apis()
    with GraphQLClient(base_url, auth=auth, timeout_seconds=30.0) as client:
        yield from iter_projects_with_opsgenie_linkable_teams(
            client,
            cloud_id,
            project_types,
            page_size,
            experimental_apis=experimental_apis or None,
        )

def _paginate_opsgenie_teams(
    *,
    client: GraphQLClient,
    cloud_id: str,
    project: api.JiraProjectNode,
    initial_connection: api.OpsgenieTeamsConnection,
    page_size: int,
    experimental_apis: Optional[Sequence[str]],
) -> List[api.OpsgenieTeamNode]:
    after = _next_after_from_pageinfo(
        has_next_page=initial_connection.page_info.has_next_page,
        end_cursor=initial_connection.page_info.end_cursor,
        edge_has_cursor=api.OPSGENIE_EDGE_HAS_CURSOR,
        edges_cursors=[e.cursor for e in initial_connection.edges],
        path=f"jira.project[{project.key}].opsgenieTeams",
    )
    if after is None:
        return []

    query = api.JIRA_PROJECT_OPSGENIE_TEAMS_PAGE_QUERY
    seen_after: set[str] = {after}
    out: List[api.OpsgenieTeamNode] = []

    while True:
        if api.REFETCH_STRATEGY == "node":
            project_id = (project.id or "").strip()
            if not project_id:
                raise SerializationError("Project id is required for node-based opsgenie pagination")
            variables = {"projectId": project_id, "first": page_size, "after": after}
        else:
            project_key = (project.key or "").strip()
            if not project_key:
                raise SerializationError("Project key is required for opsgenie pagination")
            variables = {
                "cloudId": cloud_id,
                "projectKey": project_key,
                "first": page_size,
                "after": after,
            }

        result = client.execute(
            query,
            variables=variables,
            operation_name="JiraProjectOpsgenieTeamsPage",
            experimental_apis=list(experimental_apis) if experimental_apis else None,
        )
        if result.data is None:
            raise SerializationError("Missing GraphQL data in opsgenie pagination response")

        try:
            conn = api.parse_project_opsgenie_teams(result.data)
        except SerializationError as exc:
            if result.errors:
                raise GraphQLOperationError(errors=result.errors, partial_data=result.data) from exc
            raise
        out.extend([e.node for e in conn.edges])

        next_after = _next_after_from_pageinfo(
            has_next_page=conn.page_info.has_next_page,
            end_cursor=conn.page_info.end_cursor,
            edge_has_cursor=api.OPSGENIE_EDGE_HAS_CURSOR,
            edges_cursors=[e.cursor for e in conn.edges],
            path=f"jira.project[{project.key}].opsgenieTeams",
        )
        if next_after is None:
            break
        if next_after in seen_after:
            raise SerializationError("Opsgenie pagination cursor repeated; aborting")
        seen_after.add(next_after)
        after = next_after

    return out



from __future__ import annotations

from typing import Iterator

from ...canonical_models import JiraVersion
from ...errors import SerializationError
from ..client import JiraRestClient
from ..env import auth_from_env, jira_rest_base_url_from_env
from ..gen import jira_api as api
from ..mappers.jira_versions import map_rest_version


def iter_versions_via_rest(
    client: JiraRestClient,
    project_key_or_id: str,
    page_size: int = 50,
) -> Iterator[JiraVersion]:
    project_clean = (project_key_or_id or "").strip()
    if not project_clean:
        raise ValueError("project_key_or_id is required")
    if page_size <= 0:
        raise ValueError("page_size must be > 0")

    start_at = 0
    seen_start_at: set[int] = set()

    while True:
        if start_at in seen_start_at:
            raise SerializationError(
                "Pagination startAt repeated; aborting to prevent infinite loop"
            )
        seen_start_at.add(start_at)

        payload = client.get_json(
            f"/rest/api/3/project/{project_clean}/version",
            params={"startAt": start_at, "maxResults": page_size},
        )
        page = api.PageBeanVersion.from_dict(payload, "data")
        values = page.values

        for item in values:
            yield map_rest_version(project_key=project_clean, version=item)

        has_is_last = isinstance(page.is_last, bool)
        if has_is_last and page.is_last:
            break

        has_total = isinstance(page.total, int) and page.total >= 0
        if has_total:
            if start_at + len(values) >= page.total:
                break
        else:
            if len(values) < page_size:
                break

        if len(values) == 0:
            if has_is_last and not page.is_last:
                raise SerializationError(
                    "Received empty page with isLast=false; cannot continue pagination"
                )
            break
        start_at += len(values)


def list_versions_via_rest(
    project_key_or_id: str,
    page_size: int = 50,
) -> Iterator[JiraVersion]:
    project_clean = (project_key_or_id or "").strip()
    if not project_clean:
        raise ValueError("project_key_or_id is required")

    auth = auth_from_env()
    if auth is None:
        raise ValueError("Missing credentials.")

    # Using a dummy cloud_id for URL derivation if needed, but usually derived from env
    base_url = jira_rest_base_url_from_env("")
    if not base_url:
        raise ValueError("Missing Jira REST base URL.")

    with JiraRestClient(base_url, auth=auth) as client:
        yield from iter_versions_via_rest(client, project_clean, page_size)


def create_version_via_rest(
    version: JiraVersion,
) -> JiraVersion:
    auth = auth_from_env()
    if auth is None:
        raise ValueError("Missing credentials.")

    base_url = jira_rest_base_url_from_env("")
    if not base_url:
        raise ValueError("Missing Jira REST base URL.")

    with JiraRestClient(base_url, auth=auth) as client:
        data = {
            "name": version.name,
            "project": version.project_key,
            "released": version.released,
        }
        if version.release_date:
            data["releaseDate"] = version.release_date

        payload = client.post_json("/rest/api/3/version", json_data=data)
        return map_rest_version(
            project_key=version.project_key,
            version=api.Version.from_dict(payload, "data"),
        )


def update_version_via_rest(
    version: JiraVersion,
) -> JiraVersion:
    if not version.id:
        raise ValueError("version.id is required for update")

    auth = auth_from_env()
    if auth is None:
        raise ValueError("Missing credentials.")

    base_url = jira_rest_base_url_from_env("")
    if not base_url:
        raise ValueError("Missing Jira REST base URL.")

    with JiraRestClient(base_url, auth=auth) as client:
        data = {
            "name": version.name,
            "released": version.released,
        }
        if version.release_date:
            data["releaseDate"] = version.release_date

        payload = client.put_json(f"/rest/api/3/version/{version.id}", json_data=data)
        return map_rest_version(
            project_key=version.project_key,
            version=api.Version.from_dict(payload, "data"),
        )


def delete_version_via_rest(
    version_id: str,
) -> None:
    if not version_id:
        raise ValueError("version_id is required for delete")

    auth = auth_from_env()
    if auth is None:
        raise ValueError("Missing credentials.")

    base_url = jira_rest_base_url_from_env("")
    if not base_url:
        raise ValueError("Missing Jira REST base URL.")

    with JiraRestClient(base_url, auth=auth) as client:
        client.delete(f"/rest/api/3/version/{version_id}")



from __future__ import annotations

import argparse
import os
import sys
from pathlib import Path
from typing import List, Optional
from urllib.parse import parse_qs, urlparse


def _add_project_to_syspath() -> None:
    project_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(project_root))


_add_project_to_syspath()

from atlassian.oauth_3lo import (  # noqa: E402
    build_authorize_url,
    exchange_authorization_code,
    fetch_accessible_resources,
)


def _split_scopes(raw: str) -> List[str]:
    parts: List[str] = []
    for chunk in (raw or "").replace(",", " ").split():
        v = chunk.strip()
        if v:
            parts.append(v)
    return parts


def _extract_code(user_input: str) -> str:
    raw = (user_input or "").strip()
    if not raw:
        raise ValueError("missing code/redirected URL input")
    if "://" not in raw:
        return raw
    parsed = urlparse(raw)
    qs = parse_qs(parsed.query)
    code = qs.get("code", [None])[0]
    if not code or not code.strip():
        raise ValueError("redirected URL missing ?code=")
    return code.strip()


def main(argv: Optional[List[str]] = None) -> int:
    parser = argparse.ArgumentParser(description="Atlassian OAuth 2.0 (3LO) login helper")
    parser.add_argument("--client-id", default=os.getenv("ATLASSIAN_CLIENT_ID", ""))
    parser.add_argument("--client-secret", default=os.getenv("ATLASSIAN_CLIENT_SECRET", ""))
    parser.add_argument(
        "--redirect-uri",
        default=os.getenv("ATLASSIAN_OAUTH_REDIRECT_URI", "http://localhost:8080/callback"),
    )
    parser.add_argument(
        "--scopes",
        default=os.getenv("ATLASSIAN_OAUTH_SCOPES", ""),
        help="Space- or comma-separated scopes (must match your app config)",
    )
    parser.add_argument("--state", default=os.getenv("ATLASSIAN_OAUTH_STATE", "").strip() or None)
    parser.add_argument(
        "--print-accessible-resources",
        action="store_true",
        help="After login, call accessible-resources and print cloud IDs",
    )
    args = parser.parse_args(argv)

    scopes = _split_scopes(args.scopes)
    if not args.client_id or not args.client_secret or not scopes:
        print(
            "Missing required inputs. Provide --client-id, --client-secret, and --scopes "
            "(or set ATLASSIAN_CLIENT_ID, ATLASSIAN_CLIENT_SECRET, ATLASSIAN_OAUTH_SCOPES).",
            file=sys.stderr,
        )
        return 2

    authorize_url = build_authorize_url(
        client_id=args.client_id,
        redirect_uri=args.redirect_uri,
        scopes=scopes,
        state=args.state,
    )
    print("Open this URL in your browser and complete consent:")
    print(authorize_url)
    print("", file=sys.stderr)
    print("Paste the redirected URL (or just the `code` value):", file=sys.stderr)
    try:
        raw = input().strip()
        code = _extract_code(raw)
    except (EOFError, ValueError) as exc:
        print(f"Input error: {exc}", file=sys.stderr)
        return 2

    token = exchange_authorization_code(
        client_id=args.client_id,
        client_secret=args.client_secret,
        code=code,
        redirect_uri=args.redirect_uri,
    )

    print("")
    print("# Use these in your shell or .env (do NOT commit secrets):")
    print(f"ATLASSIAN_OAUTH_ACCESS_TOKEN={token.access_token}")
    if token.refresh_token:
        print(f"ATLASSIAN_OAUTH_REFRESH_TOKEN={token.refresh_token}")
    else:
        print("# No refresh_token returned; ensure your app is configured for offline_access.")

    if args.print_accessible_resources:
        try:
            resources = fetch_accessible_resources(access_token=token.access_token)
        except Exception as exc:
            print(f"Failed to fetch accessible resources: {exc}", file=sys.stderr)
            return 0

        print("")
        print("# Accessible resources (cloud IDs):")
        for r in resources:
            rid = r.get("id")
            name = r.get("name")
            url = r.get("url")
            if isinstance(rid, str) and rid and isinstance(name, str) and isinstance(url, str):
                scopes = r.get("scopes")
                scopes_str = ""
                if isinstance(scopes, list):
                    cleaned = [s for s in scopes if isinstance(s, str) and s.strip()]
                    if cleaned:
                        scopes_str = f" scopes={','.join(cleaned)}"
                print(f"- {name}: id={rid} url={url}{scopes_str}")

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from datetime import datetime, timedelta, timezone
import json

import httpx
import pytest

from atlassian.auth import OAuthBearerAuth
from atlassian.graph.api.jira_projects import iter_projects_with_opsgenie_linkable_teams
from atlassian.graph.client import GraphQLClient


def test_jira_projects_retries_on_429_retry_after_timestamp():
    now = datetime(2021, 5, 10, 10, 59, 58, tzinfo=timezone.utc)
    current = {"now": now}

    def now_fn():
        return current["now"]

    slept: list[float] = []

    def sleeper(seconds: float) -> None:
        slept.append(seconds)
        current["now"] = current["now"] + timedelta(seconds=seconds)

    responses = [
        lambda request: httpx.Response(
            429,
            headers={"Retry-After": "2021-05-10T11:00Z"},
            json={"extensions": {"requestId": "abc-123"}},
            request=request,
        ),
        lambda request: httpx.Response(
            200,
            json={
                "data": {
                    "jira": {
                        "projects": {
                            "pageInfo": {"hasNextPage": False, "endCursor": None},
                            "edges": [],
                        }
                    }
                }
            },
            request=request,
        ),
    ]

    def handler(request: httpx.Request) -> httpx.Response:
        payload = json.loads(request.content.decode("utf-8"))
        assert payload.get("operationName") == "JiraProjectsPage"
        return responses.pop(0)(request)

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        client = GraphQLClient(
            "https://api.atlassian.com",
            auth=OAuthBearerAuth(lambda: "token"),
            max_retries_429=1,
            time_provider=now_fn,
            sleeper=sleeper,
            http_client=http_client,
        )
        results = list(
            iter_projects_with_opsgenie_linkable_teams(
                client,
                cloud_id="cloud-123",
                project_types=["SOFTWARE"],
                page_size=50,
            )
        )

    assert results == []
    assert slept and pytest.approx(slept[0], rel=0.01) == 2.0



from __future__ import annotations

from typing import Iterator

from ...canonical_models import JiraBoard
from ...errors import SerializationError
from ..client import JiraRestClient
from ..env import auth_from_env, jira_rest_base_url_from_env
from ..gen import jira_agile_api as api
from ..mappers.jira_boards import map_rest_board


def iter_boards_via_rest(
    client: JiraRestClient,
    page_size: int = 50,
) -> Iterator[JiraBoard]:
    if page_size <= 0:
        raise ValueError("page_size must be > 0")

    start_at = 0
    seen_start_at: set[int] = set()

    while True:
        if start_at in seen_start_at:
            raise SerializationError(
                "Pagination startAt repeated; aborting to prevent infinite loop"
            )
        seen_start_at.add(start_at)

        payload = client.get_json(
            "/rest/agile/1.0/board",
            params={"startAt": start_at, "maxResults": page_size},
        )
        page = api.BoardPage.from_dict(payload, "data")
        values = page.values

        for item in values:
            yield map_rest_board(board=item)

        has_is_last = isinstance(page.is_last, bool)
        if has_is_last and page.is_last:
            break

        if len(values) < page_size:
            break

        if len(values) == 0:
            break

        start_at += len(values)


def list_boards_via_rest(
    page_size: int = 50,
) -> Iterator[JiraBoard]:
    auth = auth_from_env()
    if auth is None:
        raise ValueError("Missing credentials.")

    base_url = jira_rest_base_url_from_env("")
    if not base_url:
        raise ValueError("Missing Jira REST base URL.")

    with JiraRestClient(base_url, auth=auth) as client:
        yield from iter_boards_via_rest(client, page_size)



from __future__ import annotations

import json
import os
import sys
from pathlib import Path


def _add_project_to_syspath() -> None:
    project_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(project_root))


_add_project_to_syspath()

from atlassian.auth import (  # noqa: E402
    BasicApiTokenAuth,
    CookieAuth,
    OAuthBearerAuth,
)
from atlassian.oauth_3lo import OAuthRefreshTokenAuth  # noqa: E402
from atlassian.graph.schema_fetcher import fetch_schema_introspection  # noqa: E402


def _auth_from_env():
    token = os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN")
    refresh_token = os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
    client_id = os.getenv("ATLASSIAN_CLIENT_ID")
    client_secret = os.getenv("ATLASSIAN_CLIENT_SECRET")
    email = os.getenv("ATLASSIAN_EMAIL")
    api_token = os.getenv("ATLASSIAN_API_TOKEN")
    cookies_json = os.getenv("ATLASSIAN_COOKIES_JSON")

    if refresh_token and client_id and client_secret:
        return OAuthRefreshTokenAuth(
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=refresh_token,
        )
    if token:
        if client_secret and token.strip() == client_secret.strip():
            raise ValueError(
                "ATLASSIAN_OAUTH_ACCESS_TOKEN appears to be set to ATLASSIAN_CLIENT_SECRET; "
                "set an OAuth access token (not the client secret)."
            )
        return OAuthBearerAuth(lambda: token)
    if email and api_token:
        return BasicApiTokenAuth(email, api_token)
    if cookies_json:
        try:
            cookies = json.loads(cookies_json)
        except json.JSONDecodeError:
            return None
        if isinstance(cookies, dict) and all(
            isinstance(k, str) and isinstance(v, str) for k, v in cookies.items()
        ):
            return CookieAuth(cookies)
    return None


def _experimental_apis():
    raw = os.getenv("ATLASSIAN_GQL_EXPERIMENTAL_APIS", "")
    return [p.strip() for p in raw.split(",") if p.strip()]


def main() -> int:
    repo_root = Path(__file__).resolve().parents[2]
    out_dir = repo_root / "graphql"

    base_url = os.getenv("ATLASSIAN_GQL_BASE_URL")
    if not base_url and (
        os.getenv("ATLASSIAN_OAUTH_ACCESS_TOKEN")
        or os.getenv("ATLASSIAN_OAUTH_REFRESH_TOKEN")
    ):
        base_url = "https://api.atlassian.com"
    try:
        auth = _auth_from_env()
    except ValueError as exc:
        print(str(exc), file=sys.stderr)
        return 2
    if not base_url or auth is None:
        print(
            "Missing credentials. Set ATLASSIAN_GQL_BASE_URL and one of: "
            "ATLASSIAN_OAUTH_ACCESS_TOKEN, or ATLASSIAN_OAUTH_REFRESH_TOKEN + (ATLASSIAN_CLIENT_ID + ATLASSIAN_CLIENT_SECRET), "
            "or (ATLASSIAN_EMAIL + ATLASSIAN_API_TOKEN), or ATLASSIAN_COOKIES_JSON.",
            file=sys.stderr,
        )
        return 2

    result = fetch_schema_introspection(
        base_url,
        auth,
        output_dir=out_dir,
        experimental_apis=_experimental_apis() or None,
        timeout_seconds=30.0,
    )
    print(f"Wrote {result.introspection_json_path}")
    if result.sdl_path:
        print(f"Wrote {result.sdl_path}")
    else:
        print("SDL not generated (optional dependency missing).")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

from ...canonical_models import JiraVersion
from ..gen import jira_api as api


def map_rest_version(project_key: str, version: api.Version) -> JiraVersion:
    return JiraVersion(
        id=version.id or "",
        name=version.name or "",
        project_key=project_key,
        released=version.released or False,
        release_date=version.release_date,
    )



from __future__ import annotations

from typing import Iterator

from ...canonical_models import JiraWorklog
from ...errors import SerializationError
from ..client import JiraRestClient
from ..env import auth_from_env, jira_rest_base_url_from_env
from ..gen import jira_api as api
from ..mappers.jira_worklogs import map_worklog


def iter_issue_worklogs_via_rest(
    client: JiraRestClient,
    *,
    issue_key: str,
    page_size: int = 100,
) -> Iterator[JiraWorklog]:
    issue_key_clean = (issue_key or "").strip()
    if not issue_key_clean:
        raise ValueError("issue_key is required")
    if page_size <= 0:
        raise ValueError("page_size must be > 0")

    start_at = 0
    seen_start_at: set[int] = set()

    while True:
        if start_at in seen_start_at:
            raise SerializationError(
                "Pagination startAt repeated; aborting to prevent infinite loop"
            )
        seen_start_at.add(start_at)

        payload = client.get_json(
            f"/rest/api/3/issue/{issue_key_clean}/worklog",
            params={"startAt": start_at, "maxResults": page_size},
        )
        page = api.PageOfWorklogs.from_dict(payload, "data")
        worklogs = page.worklogs

        for wl in worklogs:
            yield map_worklog(issue_key=issue_key_clean, worklog=wl)

        has_total = isinstance(page.total, int) and page.total >= 0
        if has_total:
            if start_at + len(worklogs) >= page.total:
                break
        else:
            if len(worklogs) < page_size:
                break

        if len(worklogs) == 0:
            break
        start_at += len(worklogs)


def create_worklog_via_rest(
    issue_key: str,
    time_spent_seconds: int,
    started_at: str,
) -> JiraWorklog:
    auth = auth_from_env()
    if auth is None:
        raise ValueError("Missing credentials.")

    base_url = jira_rest_base_url_from_env("")
    if not base_url:
        raise ValueError("Missing Jira REST base URL.")

    with JiraRestClient(base_url, auth=auth) as client:
        data = {
            "timeSpentSeconds": time_spent_seconds,
            "started": started_at,
        }
        payload = client.post_json(
            f"/rest/api/3/issue/{issue_key}/worklog", json_data=data
        )
        wl = api.Worklog.from_dict(payload, "data")
        return map_worklog(issue_key=issue_key, worklog=wl)


def delete_worklog_via_rest(
    issue_key: str,
    worklog_id: str,
) -> None:
    auth = auth_from_env()
    if auth is None:
        raise ValueError("Missing credentials.")

    base_url = jira_rest_base_url_from_env("")
    if not base_url:
        raise ValueError("Missing Jira REST base URL.")

    with JiraRestClient(base_url, auth=auth) as client:
        client.delete(f"/rest/api/3/issue/{issue_key}/worklog/{worklog_id}")



from pathlib import Path

import httpx

from atlassian.rest.openapi_fetcher import fetch_jira_rest_openapi


def test_fetch_jira_rest_openapi_writes_pretty_json(tmp_path: Path):
    sample = {"openapi": "3.0.1", "info": {"title": "Jira", "version": "x"}}

    def handler(request: httpx.Request) -> httpx.Response:
        assert request.method == "GET"
        return httpx.Response(200, json=sample, request=request)

    transport = httpx.MockTransport(handler)
    with httpx.Client(transport=transport, timeout=5.0) as http_client:
        out_path = tmp_path / "swagger.json"
        written = fetch_jira_rest_openapi(
            url="https://example/swagger.json",
            output_path=out_path,
            http_client=http_client,
        )

    assert written == out_path
    text = out_path.read_text(encoding="utf-8")
    assert text.endswith("\n")
    assert '"openapi": "3.0.1"' in text
